import os
import re
import sys
import shutil
import tempfile
import contextlib

import click
import yaml
from tqdm import tqdm

from shub import config as shub_config
from shub import utils as shub_utils
from shub.exceptions import (
    ShubException, NotFoundException, BadConfigException, RemoteErrorException,
    ShubDeprecationWarning, print_warning, BadParameterException,
)

if sys.version_info < (3, 10):
    import importlib_metadata as metadata
else:
    from importlib import metadata


STATUS_FILE_LOCATION = '.releases'
_VALIDSPIDERNAME = re.compile('^[a-z0-9][-._a-z0-9]+$', re.I)

DOCKER_PY_UNAVAILABLE_MSG = """\
You need docker>=2.0.0 python package installed to run the command.
docker-py python package must be uninstalled to avoid dependency conflicts.
"""
DOCKER_UNAVAILABLE_MSG = """
Detected error connecting to Docker daemon's host.

Please ensure that you have Docker installed, configured and running locally,
that's essential for running shub image command. To check that run command

    docker version

and check its output: it should contain Docker client and server versions and
should not contain any errors.

You can learn about Docker at https://www.docker.com/.
"""


def is_verbose():
    ctx = click.get_current_context(True)
    return ctx and (ctx.params.get('verbose') or ctx.params.get('debug'))


def debug_log(msg):
    if is_verbose():
        click.echo(msg)


def deprecate_debug_parameter(ctx, param, value):
    if value:
        print_warning("-d/--debug parameter is deprecated. "
                      "Please use -v/--verbose parameter instead.",
                      ShubDeprecationWarning)
    return value


def deprecate_async_parameter(ctx, param, value):
    if value:
        print_warning("--async parameter is deprecated.", ShubDeprecationWarning)
    return value


def get_project_dir():
    """ A helper to get project root dir.
        Used by init/build command to locate Dockerfile.
    """
    closest = shub_utils.closest_file('scrapinghub.yml')
    if not closest:
        raise BadConfigException(
            "Not inside a project: scrapinghub.yml not found.")
    return os.path.dirname(closest)


def get_docker_client(validate=True):
    """A helper to initiate Docker client"""
    try:
        import docker
    except ImportError:
        raise ImportError(DOCKER_PY_UNAVAILABLE_MSG)
    for dep in metadata.distributions():
        if dep.name == 'docker-py':
            raise ImportError(DOCKER_PY_UNAVAILABLE_MSG)

    docker_host = os.environ.get('DOCKER_HOST')
    tls_config = None
    if os.environ.get('DOCKER_TLS_VERIFY', False):
        tls_cert_path = os.environ.get('DOCKER_CERT_PATH')
        if not tls_cert_path:
            tls_cert_path = os.path.join(os.path.expanduser('~'), '.docker')
        apply_path_fun = lambda name: os.path.join(tls_cert_path, name)  # noqa
        tls_config = docker.tls.TLSConfig(
            client_cert=(apply_path_fun('cert.pem'),
                         apply_path_fun('key.pem')),
            verify=apply_path_fun('ca.pem'),
            assert_hostname=False)
        docker_host = docker_host.replace('tcp://', 'https://')
    version = os.environ.get('DOCKER_API_VERSION', 'auto')

    # If it returns an error, check if you have the old docker-py installed
    # together with the new docker lib, and uninstall docker-py.
    client = docker.APIClient(base_url=docker_host,
                              version=version,
                              tls=tls_config)
    if validate:
        validate_connection_with_docker_daemon(client)
    return client


def validate_connection_with_docker_daemon(client):
    try:
        client.version()
    except:  # noqa
        raise ShubException(DOCKER_UNAVAILABLE_MSG)


def format_image_name(image_name, image_tag):
    """Format image name using image tag"""
    parts = image_name.rsplit('/', 1)
    # check if tag is already here
    if ':' in parts[-1]:
        # change name to shorter version w/o existing tag
        click.echo('Please use --version param to specify tag')
        image_name = image_name.rsplit(':', 1)[0]
    if not image_tag:
        config = shub_config.load_shub_config()
        image_tag = config.get_version()
    return f'{image_name}:{image_tag}'


def get_image_registry(image_name):
    """Extract registry host from Docker image name.

    Returns None if registry hostname is not found in the name, meaning
    that default Docker registry should be used.

    Docker image name is defined according to the following rules:
     - name          := [hostname '/'] component ['/' component]*
     - hostname      := hostcomponent ['.' hostcomponent]* [':' port-number]
     - hostcomponent := /([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9])/
    https://github.com/docker/distribution/blob/master/reference/reference.go
    """
    components = image_name.split('/')
    if len(components) > 1 and any(sym in components[0] for sym in '.:'):
        return components[0]


def get_credentials(username=None, password=None, insecure=False,
                    apikey=None, target_apikey=None):
    """ A helper function to get credentials based on cmdline options.

    Returns a tuple with 2 strings: (username, password).

    When working with registries where only username matters:
    missing password leads to auth request to registry authentication service
    without 'account' query parameter which breaks login.
    """
    if insecure:
        if username or password:
            raise BadParameterException(
                'Insecure credentials not compatible with username/password')
        return None, None
    elif apikey:
        return apikey, ' '
    elif username:
        if password is None:
            raise BadParameterException(
                'Password is required when passing username.')
        return username, password
    elif password:
        raise BadParameterException(
            'Username is required when passing password.')
    return target_apikey, ' '


def store_status_url(status_url, limit):
    """Load status file and update it with a url"""
    data = _load_status_file(STATUS_FILE_LOCATION)
    if not data:
        data[0] = status_url
        _update_status_file(data, STATUS_FILE_LOCATION)
        return 0
    for stored_id, stored_url in data.items():
        if stored_url == status_url:
            click.echo(f"Found same status_url: {stored_id}")
            return stored_id
    status_id = max(data.keys()) + 1
    data[status_id] = status_url
    if len(data) > limit:
        del data[min(data.keys())]
    _update_status_file(data, STATUS_FILE_LOCATION)
    return status_id


def load_status_url(status_id):
    """ Load status url from file by status_id"""
    if not os.path.isfile(STATUS_FILE_LOCATION):
        raise NotFoundException(
            f'Status file is not found at {STATUS_FILE_LOCATION}')
    data = _load_status_file(STATUS_FILE_LOCATION)
    # return latest status url if status id is not provided
    if not isinstance(status_id, int) and data:
        max_status_id = max(data.keys())
        click.echo('Getting results for latest status id {}.'
                   .format(max_status_id))
        return data[max_status_id]
    if status_id not in data:
        raise NotFoundException(
            f"Status url with id {status_id} is not found")
    return data[status_id]


def _load_status_file(path):
    """ Open status file and parse it """
    data = {}
    if not os.path.isfile(path):
        return data
    with open(path) as f:
        try:
            data = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            raise BadConfigException(
                f"Error reading releases file:\n{exc}")
    if not isinstance(data, dict):
        raise BadConfigException(
            f"Releases file has wrong format ({data}).")
    return data


def _update_status_file(data, path):
    """ Save status file with updated data """
    with open(path, 'w') as status_file:
        yaml.dump(data, status_file, default_flow_style=False)


def valid_spiders(entries):
    """Filter out garbage and only let valid spider names in
    >>> _valid_spiders(['Update rootfs','sony.com', '', 'soa-uk', '182-blink.com'])
    ['182-blink.com', 'soa-uk', 'sony.com']
    >>> _valid_spiders(['-spiders', 'A77aque'])
    ['A77aque']
    """
    return sorted(filter(_VALIDSPIDERNAME.match, entries))


def ensure_unicode(s, encoding='utf-8'):
    return s.decode(encoding) if isinstance(s, bytes) else s


class BaseProgress:
    """Small helper class to track progress.

    Base implementation stores events iterator and walks through it with
    show() method, handle_event() logic mostly depends on operation.
    """
    def __init__(self, events):
        self.events = events

    def show(self):
        for event in self.events:
            self.handle_event(event)

    def handle_event(self, event):
        if 'error' in event:
            tqdm.write("Error {}: {}".format(event['error'],
                                             event['errorDetail']))
            raise RemoteErrorException("Docker operation failed")


class ProgressBar(tqdm):
    """Fixed version of tqdm.tqdm progress bar."""

    def moveto(self, *args, **kwargs):
        super().moveto(*args, **kwargs)
        if hasattr(self.fp, 'flush'):
            self.fp.flush()


def create_progress_bar(total, desc, **kwargs):
    """Helper creating a progress bar instance for a given parameters set.

    The bar should be closed by calling close() method.
    """
    return ProgressBar(
        total=total,
        desc=desc,
        # XXX: click.get_text_stream or click.get_binary_stream don't
        # work well with tqdm on Windows and Python 3
        file=sys.stdout,
        # helps to update bars on resizing terminal
        dynamic_ncols=True,
        # miniters improves progress on erratic updates caused by network
        miniters=1,
        **kwargs
    )


@contextlib.contextmanager
def make_temp_directory(cleanup=True, **kwargs):
    temp_dir = tempfile.mkdtemp(**kwargs)
    try:
        yield temp_dir
    finally:
        if cleanup:
            shutil.rmtree(temp_dir)



import click

from scrapinghub import ScrapinghubAPIError
from scrapinghub.client.utils import parse_job_key

from shub.utils import get_scrapinghub_client_from_config
from shub.config import get_target_conf
from shub.exceptions import (
    ShubException,
    BadParameterException,
    SubcommandException,
)


HELP = """
Cancel multiple jobs from Scrapy Cloud.

The cancel command expects the project ID (target) followed by
the pair containing the spider ID and Job ID:

    shub cancel 12345 1/1 1/2 1/3

If the project ID is not defined it is going to use the default
project (as defined in scrapinghub.yml):

    shub cancel 1/1 1/2 1/3

The cancel command requires a confirmation that could be skipped
with the flag --force/-f:

    shub cancel --force 1/1 1/2 1/3
"""


SHORT_HELP = "Cancel multiple jobs from Scrapy Cloud"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target_or_key")
@click.argument("keys", nargs=-1)
@click.option('--force', '-f', is_flag=True,
              help='It ignores the confirmation prompt')
def cli(target_or_key, keys, force):
    # target_or_key contains a target or just another job key
    if "/" in target_or_key:
        keys = (target_or_key,) + keys
        target = "default"
    else:
        target = target_or_key

    targetconf = get_target_conf(target)
    project_id = targetconf.project_id
    client = get_scrapinghub_client_from_config(targetconf)
    project = client.get_project(project_id)

    try:
        job_keys = [validate_job_key(project_id, key) for key in keys]
    except (BadParameterException, SubcommandException) as err:
        click.echo('Error during keys validation: %s' % str(err))
        exit(1)

    if not force:
        jobs_str = ", ".join([str(job) for job in job_keys])
        click.confirm(
            'Do you want to cancel these %s jobs? \n\n%s \n\nconfirm?'
            % (len(job_keys), jobs_str),
            abort=True
        )

    try:
        output = project.jobs.cancel(
            keys=[str(job) for job in job_keys]
        )
    except (ValueError, ScrapinghubAPIError) as err:
        raise ShubException(str(err))

    click.echo(output)


def validate_job_key(project_id, short_key):
    job_key = f"{project_id}/{short_key}"

    if len(short_key.split("/")) != 2:
        raise BadParameterException(
            "keys must be defined as <spider_id>/<job_id>"
        )

    try:
        return parse_job_key(job_key)
    except ValueError as err:
        raise BadParameterException(str(err))
    except Exception as err:
        raise SubcommandException(str(err))



import unittest
import os
import tempfile
from unittest import mock

from click.testing import CliRunner

from shub import deploy_reqs

from .utils import mock_conf


class TestDeployReqs(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.conf = mock_conf(self)

    @unittest.skip('flaky')
    def test_can_decompress_downloaded_packages_and_call_deploy_reqs(self):
        requirements_file = self._write_tmp_requirements_file()
        with mock.patch('shub.utils.build_and_deploy_egg') as m:
            self.runner.invoke(
                deploy_reqs.cli,
                ('-r', requirements_file),
            )
            self.assertEqual(m.call_count, 2)
            for args, kwargs in m.call_args_list:
                project, endpoint, apikey = args
                self.assertEqual(project, 1)
                self.assertIn('https://app.zyte.com', endpoint)
                self.assertEqual(apikey, self.conf.apikeys['default'])

    def _write_tmp_requirements_file(self):
        basepath = 'tests/samples/deploy_reqs_sample_project/'
        eggs = ['other-egg-0.2.1.zip', 'inflect-0.2.5.tar.gz']
        tmp_dir = tempfile.mkdtemp(prefix="shub-test-deploy-reqs")
        requirements_file = os.path.join(tmp_dir, 'requirements.txt')

        with open(requirements_file, 'w') as f:
            for egg in eggs:
                f.write(os.path.abspath(os.path.join(basepath, egg)) + "\n")

        return requirements_file



import glob
import json
import os
import shutil
import tempfile
from typing import AnyStr, Optional, Union

# Not used in code but needed in runtime, don't remove!
import setuptools
import setuptools.msvc  # noqa

import click
import toml
from urllib.parse import urljoin

from shub.config import SH_IMAGES_REGISTRY, list_targets_callback, load_shub_config
from shub.exceptions import BadParameterException, NotFoundException, ShubException, SubcommandException
from shub.image.upload import upload_cmd
from shub.utils import (create_default_setup_py, create_scrapinghub_yml_wizard,
                        inside_project, make_deploy_request, run_cmd, run_python)

HELP = """
Deploy the current folder's Scrapy project to Scrapy Cloud.

If you do not supply `target`, the default target from scrapinghub.yml will be
used. If you have no scrapinghub.yml, you will be guided through a short wizard
to create one. You can also specify a numerical project ID:

    shub deploy 12345

Or use any of the targets defined in your scrapinghub.yml:

    shub deploy production

To see a list of all defined targets, run:

    shub deploy -l

You can also deploy an existing project egg:

    shub deploy --egg egg_name

Or build an egg without deploying:

    shub deploy --build-egg egg_name
"""

SHORT_HELP = "Deploy Scrapy project to Scrapy Cloud"


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-V", "--version", help="The version to use for deploying")
@click.option("-d", "--debug", help="Debug mode (do not remove build dir)",
              is_flag=True)
@click.option("--egg", help="Deploy the given egg, instead of building one")
@click.option("--build-egg", help="Only build the given egg, don't deploy it")
@click.option("-v", "--verbose", help="Stream deploy logs to console",
              is_flag=True)
@click.option("-k", "--keep-log", help="Keep the deploy log", is_flag=True)
@click.option("--ignore-size", help="Ignore deploy request's egg(s) size check",
              is_flag=True)
def cli(target, version, debug, egg, build_egg, verbose, keep_log,
        ignore_size):
    conf, image = load_shub_config(), None
    if not build_egg:
        create_scrapinghub_yml_wizard(conf, target=target)
    image = conf.get_target_conf(target).image
    if not image:
        deploy_cmd(target, version, debug, egg, build_egg, verbose, keep_log,
                   conf=conf)
    elif image.startswith(SH_IMAGES_REGISTRY):
        upload_cmd(target, version)
    else:
        raise BadParameterException(
            "Please use `shub image` commands to work with Docker registries "
            "other than Scrapinghub default registry.")


def deploy_cmd(target, version, debug, egg, build_egg, verbose, keep_log,
               conf=None):
    tmpdir = None
    try:
        if build_egg:
            egg, tmpdir = _build_egg()
            click.echo("Writing egg to %s" % build_egg)
            shutil.copyfile(egg, build_egg)
        else:
            conf = conf or load_shub_config()
            targetconf = conf.get_target_conf(target)
            version = version or targetconf.version
            auth = (targetconf.apikey, '')

            if egg:
                click.echo("Using egg: %s" % egg)
                egg = egg
            else:
                click.echo("Packing version %s" % version)
                egg, tmpdir = _build_egg()

            _upload_egg(targetconf.endpoint, egg, targetconf.project_id,
                        version, auth, verbose, keep_log, targetconf.stack,
                        targetconf.requirements_file, targetconf.eggs, tmpdir)
            click.echo("Run your spiders at: "
                       "https://app.zyte.com/p/%s/"
                       "" % targetconf.project_id)
    finally:
        if tmpdir:
            if debug:
                click.echo("Output dir not removed: %s" % tmpdir)
            else:
                shutil.rmtree(tmpdir, ignore_errors=True)


def _url(endpoint, action):
    return urljoin(endpoint, action)


def _upload_egg(endpoint, eggpath, project, version, auth, verbose, keep_log,
                stack=None, requirements_file=None, eggs=None, tmpdir=None):
    expanded_eggs = []
    for e in (eggs or []):
        # Expand glob patterns, but make sure we don't swallow non-existing
        # eggs that were directly named
        # (glob.glob('non_existing_file') returns [])
        if any(['*' in e, '?' in e, '[' in e and ']' in e]):
            # Never match the main egg
            expanded_eggs.extend(
                [x for x in glob.glob(e)
                 if os.path.abspath(x) != os.path.abspath(eggpath)])
        else:
            expanded_eggs.append(e)

    data = {'project': project, 'version': version}
    if stack:
        data['stack'] = stack

    try:
        files = [('eggs', open(path, 'rb')) for path in expanded_eggs]
        if _is_pipfile(requirements_file):
            requirements_file = _get_pipfile_requirements(tmpdir)
        elif _is_poetry(requirements_file):
            requirements_file = _get_poetry_requirements()
        elif requirements_file:
            requirements_file = open(requirements_file, 'rb')
        if requirements_file:
            files.append(('requirements', requirements_file))
    except OSError as e:
        raise ShubException(f"{e.strerror} {e.filename}")
    files.append(('egg', open(eggpath, 'rb')))
    url = _url(endpoint, 'scrapyd/addversion.json')
    click.echo('Deploying to Scrapy Cloud project "%s"' % project)
    return make_deploy_request(url, data, files, auth, verbose, keep_log)


def _is_pipfile(name):
    return name in ['Pipfile', 'Pipfile.lock']


def _get_pipfile_requirements(tmpdir=None):
    try:
        # moved in pipenv==2022.4.8
        from pipenv.utils.dependencies import convert_deps_to_pip
        from pipenv.utils.indexes import prepare_pip_source_args
    except ImportError:
        try:
            from pipenv.utils import convert_deps_to_pip, prepare_pip_source_args
        except ImportError:
            raise ImportError('You need pipenv installed to deploy with Pipfile')
    try:
        with open('Pipfile.lock') as f:
            pipefile = json.load(f)
            deps = pipefile['default']
            sources_list = prepare_pip_source_args(pipefile['_meta']['sources'])
            sources = ' '.join(sources_list)
    except OSError:
        raise ShubException('Please lock your Pipfile before deploying')
    # We must remove any hash from the pipfile before converting to play nice
    # with vcs packages
    for k, v in deps.items():
        if 'hash' in v:
            del v['hash']
        if 'hashes' in v:
            del v['hashes']
        # Scrapy Cloud also doesn't support editable packages
        if 'editable' in v:
            del v['editable']
    return open(_add_sources(convert_deps_to_pip(deps), _sources=sources.encode(), tmpdir=tmpdir), 'rb')


def _add_sources(
    _requirements: Union[str, list], _sources: bytes, tmpdir: Optional[AnyStr] = None
) -> str:
    tmp = tempfile.NamedTemporaryFile(delete=False, suffix="-requirements.txt", dir=tmpdir)
    tmp.write(_sources + b'\n')
    # Keep backward compatibility with pipenv<=2022.8.30
    if isinstance(_requirements, list):
        tmp.write('\n'.join(_requirements).encode('utf-8'))
    # Keep compatible with pipenv>=v2023.10.24
    elif isinstance(_requirements, dict):
        tmp.write('\n'.join(_requirements.values()).encode('utf-8'))
    else:
        with open(_requirements, 'rb') as f:
            tmp.write(f.read())
    tmp.flush()
    tmp.close()
    return tmp.name


def _is_poetry(name):
    if name != 'pyproject.toml':
        return False
    data = toml.load(name)
    return 'poetry' in (data.get('tool') or {})


def _get_poetry_requirements_fallback():
    data = toml.load('poetry.lock')
    # Adapted from poetry 1.0.0a2 poetry/utils/exporter.py
    lines = []
    for package in data['package']:
        source = package.get('source') or {}
        source_type = source.get('type')
        if source_type == 'git':
            line = 'git+{}@{}#egg={}'.format(
                source['url'], source['reference'], package['name']
            )
        elif source_type in ['directory', 'file']:
            line = ''
            line += source['url']
        else:
            line = '{}=={}'.format(package['name'], package['version'])

            if source_type == 'legacy' and source['url']:
                line += ' \\\n'
                line += '    --index-url {}'.format(source['url'])

        line += '\n'
        lines.append(line)
    return ''.join(lines)


def _get_poetry_requirements():
    executable = shutil.which("poetry")
    if executable is None:
        try:
            return _get_poetry_requirements_fallback()
        except Exception:
            raise NotFoundException("poetry executable not found.")
    try:
        return run_cmd([executable, "export", "--without-hashes", "-f", "requirements.txt"])
    except SubcommandException as original_exception:
        try:
            return _get_poetry_requirements_fallback()
        except Exception:
            raise original_exception


def _build_egg():
    if not inside_project():
        raise NotFoundException("No Scrapy project found in this location.")
    create_default_setup_py()
    d = tempfile.mkdtemp(prefix="shub-deploy-")
    run_python(['setup.py', 'clean', '-a', 'bdist_egg', '-d', d])
    egg = glob.glob(os.path.join(d, '*.egg'))[0]
    return egg, d



#!/usr/bin/env python
"""
A helper wrapper over start-crawl to run a custom image locally.

The wrapper is used in `shub image run` command as an entrypoint
to create a FIFO file inside a Docker container, enforce using it
to communicate with crawl process and start the crawl process.

The initial version handles and prints only LOG entries to mimic
Scrapy behavior when running locally, however it could be easily
extended in the future.

Reading about SH custom image contract should bring you more context
https://shub.readthedocs.io/en/stable/custom-images-contract.html.

FIFO based communication protocol is described well in
https://doc.scrapinghub.com/scrapy-cloud-write-entrypoint.html

TODO As a custom image isn't necessarily based on Python, the wrapper
should be rewritten in the future with something more basic and
lightweight, to get rid of dependence on Python.
"""


import os
import sys
import json
import logging
import datetime
from multiprocessing import Process
from shutil import which


def _consume_from_fifo(fifo_path):
    """Start reading/printing entries from FIFO."""
    with open(fifo_path) as fifo:
        while True:
            line = fifo.readline()
            # returns an empty string only in the end of the file
            if not line:
                return
            entry_type, entry_raw = line[:3], line[4:]
            _print_fifo_entry(entry_type, json.loads(entry_raw))


def _print_fifo_entry(message_type, message):
    """Print only specific entries."""
    if message_type == 'LOG':
        timestamp = _millis_to_str(message['time'])
        loglevel = logging.getLevelName(message['level'])
        # mimic Scrapy logging format as much as possible
        print('{} {} {}'.format(timestamp, loglevel, message['message']))


def _millis_to_str(millis):
    """Convert a datatime in ms to a formatted string."""
    datetime_ts = datetime.datetime.fromtimestamp(millis / 1000.0)
    return datetime_ts.strftime('%Y-%m-%d %H:%M:%S')


def main():
    """Main wrapper entrypoint."""
    # create a named pipe for communication
    fifo_path = os.environ.get('SHUB_FIFO_PATH')
    os.mkfifo(fifo_path)
    # create and start a consumer process to read from the fifo:
    # non-daemon to allow it to finish reading from pipe before exit.
    Process(target=_consume_from_fifo, args=[fifo_path]).start()
    # replace current process with original start-crawl
    os.execv(which('start-crawl'), sys.argv)


if __name__ == '__main__':
    sys.exit(main())



import unittest
from click.testing import CliRunner
from shub import tool
import os


@unittest.skipUnless(os.getenv('USING_TOX'),
                     'End to end tests only run via TOX')
class ShubEndToEndTests(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()

    def run_subcmd(self, subcmd):
        return self.runner.invoke(tool.cli, [subcmd]).output

    def test_usage_is_displayed_if_no_arg_is_provided(self):
        output = self.run_subcmd('')
        usage_is_displayed = output.startswith('Usage:')
        self.assertTrue(usage_is_displayed)

    def test_deploy_egg_isnt_broken(self):
        output = self.run_subcmd('deploy-egg')
        error = 'Unexpected output: %s' % output
        self.assertTrue('specify target' in output, error)

    def test_deploy_reqs_isnt_broken(self):
        output = self.run_subcmd('deploy-reqs')
        error = 'Unexpected output: %s' % output
        self.assertTrue('specify target' in output, error)

    def test_deploy_isnt_broken(self):
        output = self.run_subcmd('deploy')
        error = 'Unexpected output: %s' % output
        self.assertTrue('Cannot find project' in output, error)

    def test_fetch_eggs_isnt_broken(self):
        output = self.run_subcmd('fetch-eggs')
        error = 'Unexpected output: %s' % output
        self.assertTrue('specify target' in output, error)



import os
import shutil
import tempfile
import textwrap
import unittest
from unittest import mock
from io import StringIO

import yaml
from click.testing import CliRunner
from yaml import CLoader as Loader

from shub.config import (get_target, get_target_conf, get_version,
                         load_shub_config, ShubConfig, Target,
                         SH_IMAGES_REPOSITORY)
from shub.exceptions import (BadParameterException, BadConfigException,
                             ConfigParseException, MissingAuthException,
                             NotFoundException)


VALID_YAML_CFG = """
    projects:
        shproj: 123
        externalproj: external/123
        notmeproj:
            id: 234
            apikey: otheruser
        advanced_prod:
            id: 456
            stack: hworker:v1.0.0
        advanced_dev:
            id: 457
            stack: dev
        some:
            id: 345
            requirements:
                file: requirements-dev.txt
                eggs:
                - ./egg3.egg
                - ./egg4.egg
    endpoints:
        external: ext_endpoint
    apikeys:
        default: key
        otheruser: otherkey
    stacks:
        dev: scrapy:v1.1
    images:
        dev: registry/user/project
        prod: user/project
    requirements:
        eggs:
          - ./egg1.egg
          - ./egg2.egg
        file: requirements.txt
    version: 1.0
"""


def _project_dict(proj_id, endpoint='default', extra=None):
    projd = {
        'id': proj_id,
        'endpoint': endpoint,
        'apikey': endpoint,
    }
    projd.update(extra or {})
    return projd


def _target(id, endpoint=None, apikey=None, stack=None, image=None,
            requirements_file='requirements.txt', version='1.0', eggs=None):
    if eggs is None:
        eggs = ['./egg1.egg', './egg2.egg']
    return Target(
        project_id=id,
        endpoint=endpoint or ShubConfig.DEFAULT_ENDPOINT,
        apikey=apikey,
        stack=stack,
        image=image,
        requirements_file=requirements_file,
        version=version,
        eggs=eggs
    )


class ShubConfigTest(unittest.TestCase):

    def _get_conf_with_yml(self, yml):
        conf = ShubConfig()
        conf.load(StringIO(textwrap.dedent(yml)))
        return conf

    def setUp(self):
        self.conf = self._get_conf_with_yml(VALID_YAML_CFG)

    def test_init_sets_default(self):
        conf = ShubConfig()
        self.assertIn('default', conf.endpoints)
        self.assertEqual(conf.version, 'AUTO')

    def test_load(self):
        projects = {
            'shproj': 123,
            'externalproj': 'external/123',
            'notmeproj': {
                'id': 234,
                'apikey': 'otheruser',
            },
            'advanced_prod': {
                'id': 456,
                'stack': 'hworker:v1.0.0',
            },
            'advanced_dev': {
                'id': 457,
                'stack': 'dev'
            },
            'some': {
                'id': 345,
                'requirements': {
                    'file': 'requirements-dev.txt',
                    'eggs': ['./egg3.egg', './egg4.egg']
                }
            }
        }
        self.assertEqual(projects, self.conf.projects)
        endpoints = {'external': 'ext_endpoint'}
        self.assertLessEqual(endpoints.items(), self.conf.endpoints.items())
        apikeys = {'default': 'key', 'otheruser': 'otherkey'}
        self.assertEqual(apikeys, self.conf.apikeys)
        stacks = {'dev': 'scrapy:v1.1'}
        self.assertEqual(stacks, self.conf.stacks)
        images = {'dev': 'registry/user/project',
                  'prod': 'user/project'}
        self.assertEqual(images, self.conf.images)
        self.assertEqual('requirements.txt', self.conf.requirements_file)

    def test_load_partial(self):
        yml = """
            endpoints:
                external: ext_endpoint
            extra:
                not related to ShubConfig
        """
        conf = self._get_conf_with_yml(yml)
        endpoints = {'external': 'ext_endpoint'}
        self.assertLessEqual(endpoints.items(), conf.endpoints.items())
        self.assertEqual(conf.projects, {})
        self.assertEqual(conf.apikeys, {})
        self.assertEqual(conf.images, {})
        # Assert no exception raised on empty file
        conf = self._get_conf_with_yml("")

    def test_load_shortcuts(self):
        yml = """
            project: 111
            stack: awesome_stack
            apikey: awesome_key
            endpoint: awesome_endpoint
            image: awesome_image
        """
        conf = self._get_conf_with_yml(yml)
        self.assertEqual(conf.projects['default'], 111)
        self.assertEqual(conf.stacks['default'], 'awesome_stack')
        self.assertEqual(conf.apikeys['default'], 'awesome_key')
        self.assertEqual(conf.endpoints['default'], 'awesome_endpoint')
        self.assertEqual(conf.images['default'], 'awesome_image')
        self.assertEqual(conf.get_target_conf('default').stack,
                         'awesome_stack')

    def test_load_shortcut_mixed(self):
        # Weird, but allowed
        yml = """
            stacks:
                dev: dev_stack
            stack: prod_stack
        """
        self.assertLessEqual(
            self._get_conf_with_yml(yml).stacks.items(),
            {'default': 'prod_stack', 'dev': 'dev_stack'}.items()
        )

    def test_load_shortcut_conflict(self):
        outer_yml = """
            project: 111
            stacks:
                default: outer
        """
        inner_yml = """
            stack: inner
        """
        conflicting_yml = """
            stacks:
                default: A
            stack: B
        """
        # It is allowed to override a default option via shortcut if it is from
        # a previously loaded file
        self.conf.load(outer_yml)
        self.conf.load(inner_yml)
        # But not to specify both stacks: default and stack
        with self.assertRaises(BadConfigException):
            self.conf.load(conflicting_yml)

    def test_load_malformed(self):
        # Invalid YAML
        yml = """
            endpoints
                external: ext_endpoint
            apikeys:
                default: key
        """
        with self.assertRaises(ConfigParseException):
            self._get_conf_with_yml(yml)
        # Valid YAML but not dictionary-like
        yml = """
            endpoints
                external: ext_endpoint
        """
        with self.assertRaises(ConfigParseException):
            self._get_conf_with_yml(yml)

    def test_load_file(self):
        tmpdir = tempfile.mkdtemp()
        tmpfilepath = os.path.join(tmpdir, 'scrapinghub.yml')
        with open(tmpfilepath, 'w') as f:
            f.write(textwrap.dedent(
                """
                apikeys:
                    external: ext_endpoint
                """
            ))
        conf = ShubConfig()
        conf.load_file(tmpfilepath)
        shutil.rmtree(tmpdir)
        self.assertEqual({'external': 'ext_endpoint'}, conf.apikeys)

    def test_load_scrapycfg(self):
        tmpdir = tempfile.mkdtemp()
        tmpfilepath = os.path.join(tmpdir, 'scrapy.cfg')

        def _get_conf(scrapycfg_default_target):
            with open(tmpfilepath, 'w') as f:
                f.write(textwrap.dedent(scrapycfg_default_target))
                f.write(textwrap.dedent(
                    """
                    [deploy:prod]
                    project = 222

                    [deploy:otheruser]
                    project = 333
                    username = otherkey

                    [deploy:otherurl]
                    project = 444
                    url = http://app.zyte.com/api/scrapyd/

                    [deploy:external]
                    project = 555
                    url = external_endpoint
                    username = externalkey

                    [deploy:invalid_external]
                    project = non-numeric
                    url = external_endpoint
                    username = externalkey
                    """
                ))
            conf = ShubConfig()
            conf.load_scrapycfg([tmpfilepath])
            return conf

        expected_projects = {
            'prod': '222',
            'otheruser': {
                'id': '333',
                'apikey': 'otheruser',
            },
            'otherurl': 'otherurl/444',
            'external': 'external/555',
        }
        expected_endpoints = {
            'default': ShubConfig.DEFAULT_ENDPOINT,
            'external': 'external_endpoint',
            'otherurl': 'http://app.zyte.com/api/'
        }
        expected_apikeys = {
            'otheruser': 'otherkey',
            'external': 'externalkey',
        }

        def _test_conf(scrapycfg_default_target):
            conf = _get_conf(scrapycfg_default_target)
            self.assertEqual(conf.projects, expected_projects)
            self.assertEqual(conf.endpoints, expected_endpoints)
            self.assertEqual(conf.apikeys, expected_apikeys)

        # Default with invalid project
        _test_conf(
            """
            [deploy]
            project = non-numeric
            """
        )

        # Default with valid project
        expected_projects['default'] = '111'
        _test_conf(
            """
            [deploy]
            project = 111
            """
        )

        # Default with URL
        del expected_projects['default']
        expected_endpoints['default'] = 'http://default_url'
        _test_conf(
            """
            [deploy]
            url = http://default_url
            """
        )

        # Default with key
        expected_endpoints['default'] = ShubConfig.DEFAULT_ENDPOINT
        expected_apikeys['default'] = 'key'
        expected_apikeys['otherurl'] = 'key'
        _test_conf(
            """
            [deploy]
            username = key
            """
        )

        shutil.rmtree(tmpdir)

    def test_deprecated_reqfile_syntax(self):
        self.assertEqual(self.conf.requirements_file, 'requirements.txt')
        self.conf.load("""
            requirements_file: oldstyle.txt
            """)
        self.assertEqual(self.conf.requirements_file, 'oldstyle.txt')
        self.conf.load("""
            requirements:
                file: newstyle.txt
            requirements_file: oldstyle.txt
            """)
        self.assertEqual(self.conf.requirements_file, 'newstyle.txt')

    def test_save(self):
        with CliRunner().isolated_filesystem():
            self.conf.save('conf.yml')
            loaded_conf = ShubConfig()
            loaded_conf.load_file('conf.yml')
            for option in ['projects', 'endpoints', 'apikeys', 'version',
                           'stacks', 'requirements_file', 'eggs', 'images']:
                self.assertEqual(
                    getattr(self.conf, option), getattr(loaded_conf, option))

    def test_save_partial(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            conf = self._get_conf_with_yml("""
                projects:
                    default: 123
                """)
            conf.save('conf.yml')
            with open('conf.yml') as f:
                self.assertEqual(yaml.load(f, Loader=Loader), {'project': 123})

            conf = self._get_conf_with_yml("""
                projects:
                    default: 123
                requirements:
                    file: reqs.txt
                """)
            conf.save('conf.yml')
            with open('conf.yml') as f:
                self.assertEqual(yaml.load(f, Loader=Loader), {
                    'project': 123,
                    'requirements': {'file': 'reqs.txt'}}
                )

    def test_save_skip_defaults(self):
        conf = ShubConfig()
        with CliRunner().isolated_filesystem():
            conf.save('conf.yml')
            with open('conf.yml') as f:
                self.assertEqual(yaml.load(f, Loader=Loader), None)

    def test_save_shortcut(self):
        conf = ShubConfig()
        conf.endpoints['ext'] = 'external'
        conf.stacks['default'] = 'my_stack'
        expected_yml_dict = {
            # No shortcut
            'endpoints': {
                'default': ShubConfig.DEFAULT_ENDPOINT,
                'ext': 'external',
            },
            # Shortcut
            'stack': 'my_stack',
        }
        with CliRunner().isolated_filesystem():
            conf.save('conf.yml')
            with open('conf.yml') as f:
                self.assertEqual(yaml.load(f, Loader=Loader), expected_yml_dict)

    def test_save_shortcut_updated(self):
        OLD_YML = """\
        projects:
            default: 12345
            prod: 33333
        """
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('scrapinghub.yml', 'w') as f:
                f.write(textwrap.dedent(OLD_YML))
            conf = ShubConfig()
            conf.load_file('scrapinghub.yml')
            del conf.projects['prod']
            print(conf.projects)
            conf.save('scrapinghub.yml')
            with open('scrapinghub.yml') as f:
                new_yml = yaml.safe_load(f)
            # Should not contain 'projects'
            self.assertEqual(new_yml, {'project': 12345})

            conf = ShubConfig()
            conf.load_file('scrapinghub.yml')
            # Should also work in reverse
            conf.projects['prod'] = 33333
            conf.save('scrapinghub.yml')
            with open('scrapinghub.yml') as f:
                new_yml = yaml.safe_load(f)
            # Should not contain 'project' singleton
            self.assertEqual(
                new_yml,
                {'projects': {'default': 12345, 'prod': 33333}},
            )

            # Make sure it is readable again
            ShubConfig().load_file('scrapinghub.yml')

    def test_save_partial_options(self):
        OLD_YML = """\
        projects:
            default: 12345
            prod: 33333
        stack: custom-stack
        """
        with CliRunner().isolated_filesystem():
            with open('conf.yml', 'w') as f:
                f.write(textwrap.dedent(OLD_YML))
            conf = ShubConfig()
            conf.load_file('conf.yml')
            del conf.projects['prod']
            del conf.stacks['default']
            conf.save('conf.yml', options=['projects'])
            with open('conf.yml') as f:
                self.assertEqual(
                    yaml.load(f, Loader=Loader),
                    {'project': 12345, 'stack': 'custom-stack'})
            conf.save('conf.yml')
            with open('conf.yml') as f:
                self.assertEqual(yaml.load(f, Loader=Loader), {'project': 12345})

    def test_normalized_projects(self):
        expected_projects = {
            'shproj': _project_dict(123),
            'externalproj': _project_dict(123, 'external'),
            'notmeproj': _project_dict(234, extra={'apikey': 'otheruser'}),
            'advanced_prod': _project_dict(
                456, extra={'stack': 'hworker:v1.0.0'}),
            'advanced_dev': _project_dict(457, extra={'stack': 'dev'}),
            'some': _project_dict(
                345, extra={
                    'requirements': {
                        'file': 'requirements-dev.txt',
                        'eggs': ['./egg3.egg', './egg4.egg']
                    }
                }
            )
        }
        self.assertEqual(self.conf.normalized_projects, expected_projects)

    def test_get_project(self):
        self.assertEqual(self.conf.get_project(123),
                         self.conf.get_project('shproj'))
        self.assertEqual(self.conf.get_project(456),
                         self.conf.get_project('advanced_prod'))
        self.assertEqual(self.conf.get_project(456),
                         self.conf.get_project('456'))
        self.assertEqual(self.conf.get_project('externalproj'),
                         self.conf.get_project('external/123'))

    def test_get_image(self):
        self.conf.load("""
            projects:
                default: 123
                stacks-explicit:
                    id: 322
                    image: false
                devel:
                    id: 456
                    image: true
                sh-alias:
                    id: 654
                    image: scrapinghub
                custom:
                    id: 789
                    image: user/repo
                deprecated: 987
            image: true
            images:
                deprecated: old/style
        """)
        self.assertEqual(self.conf.get_image('default'),
                         SH_IMAGES_REPOSITORY.format(project=123))
        with self.assertRaises(BadConfigException):
            self.conf.get_image('stacks-explicit')
        # check that aliases work as expected
        self.assertEqual(self.conf.get_image('devel'),
                         SH_IMAGES_REPOSITORY.format(project=456))
        self.assertEqual(self.conf.get_image('sh-alias'),
                         SH_IMAGES_REPOSITORY.format(project=654))
        # check if custom image is respected
        self.assertEqual(self.conf.get_image('custom'), 'user/repo')
        # check for backward compatibility
        self.assertEqual(self.conf.get_image('deprecated'), 'old/style')

    def test_get_image_not_found(self):
        self.conf.load("""
            projects:
                default: 123
        """)
        with self.assertRaises(NotFoundException):
            self.conf.get_image('default')

    def test_get_image_mismatch_project(self):
        self.conf.load("""
            projects:
                prod: 123
                develop: 321
                success: 456
            images:
                prod: images.scrapinghub.com/wrong-prefix/123
                develop: images.scrapinghub.com/project/123
                success: images.scrapinghub.com/project/456
        """)
        with self.assertRaises(BadConfigException):
            self.conf.get_image('prod')
        with self.assertRaises(BadConfigException):
            self.conf.get_image('develop')
        self.assertEqual(self.conf.get_image('success'),
                         SH_IMAGES_REPOSITORY.format(project=456))

    def test_get_image_ambigious_deprecated_images_section(self):
        with self.assertRaises(BadConfigException):
            self.conf.load("""
                projects:
                    default:
                        id: 123
                    stacks:
                        id: 322
                        stack: scrapy:1.2
                images:
                    default: custom/image
            """)
        self.conf.load("""
            projects:
                default:
                    id: 123
                stacks:
                    id: 322
                    image: false
                    stack: scrapy:1.2
            image: custom/image
        """)

    def test_get_image_ambiguous_global_image_and_global_stack(self):
        self.conf.load("""
            project: 123
            image: true
            stack: scrapy:1.3
        """)
        with self.assertRaisesRegex(BadConfigException, '(?i)ambiguous'):
            self.conf.get_image('default')

    def test_get_image_ambiguous_global_image_and_project_stack(self):
        self.conf.load("""
            projects:
              bad:
                id: 123
                stack: scrapy:1.3
              good:
                id: 456
                image: false
                stack: scrapy:1.3
            image: true
            """)
        with self.assertRaisesRegex(BadConfigException, '(?i)ambiguous'):
            self.conf.get_image('bad')
        with self.assertRaisesRegex(BadConfigException, '(?i)disabled'):
            self.conf.get_image('good')

    def test_get_image_ambiguous_project_image_and_project_stack(self):
        self.conf.load("""
            projects:
              default:
                id: 123
                image: true
                stack: scrapy:1.3
            """)
        with self.assertRaisesRegex(BadConfigException, '(?i)ambiguous'):
            self.conf.get_image('default')

    def test_get_target_conf(self):
        self.assertEqual(
            self.conf.get_target_conf('shproj', auth_required=False),
            _target(123, apikey='key')
        )
        self.assertEqual(
            self.conf.get_target_conf('shproj', auth_required=True),
            self.conf.get_target_conf('shproj', auth_required=False),
        )
        with self.assertRaises(MissingAuthException):
            self.conf.get_target_conf('externalproj')
        self.assertEqual(
            self.conf.get_target_conf('externalproj', auth_required=False),
            _target(123, 'ext_endpoint')
        )
        self.assertEqual(
            self.conf.get_target_conf('notmeproj'),
            _target(234, apikey='otherkey'),
        )
        self.assertEqual(
            self.conf.get_target_conf('advanced_prod'),
            _target(456, apikey='key', stack='hworker:v1.0.0'),
        )
        self.assertEqual(
            self.conf.get_target_conf('advanced_dev'),
            _target(457, apikey='key', stack='scrapy:v1.1'),
        )

    def test_get_target_conf_custom_defaults(self):
        self.conf.load("""
            stacks:
              default: custom_default
        """)
        self.assertEqual(
            self.conf.get_target_conf('shproj'),
            _target(123, apikey='key', stack='custom_default')
        )
        self.assertEqual(
            self.conf.get_target_conf('advanced_prod'),
            _target(456, apikey='key', stack='hworker:v1.0.0'),
        )

    def test_get_target_conf_calls_get_project(self):
        t = _target(456, apikey='key', stack='hworker:v1.0.0')
        self.assertEqual(self.conf.get_target_conf('advanced_prod'), t)
        self.assertEqual(self.conf.get_target_conf(456), t)
        self.assertEqual(self.conf.get_target_conf('456'), t)
        self.assertEqual(self.conf.get_target_conf('default/456'), t)

    def test_get_target_correct_requirements(self):
        correct_reqs = {
            'shproj': 'requirements.txt',
            'notmeproj': 'requirements.txt',
            'advanced_prod': 'requirements.txt',
            'advanced_dev': 'requirements.txt',
            'some': 'requirements-dev.txt'
        }
        for name, reqs_file in correct_reqs.items():
            target = self.conf.get_target_conf(name)
            self.assertEqual(target.requirements_file, reqs_file)

    def test_get_target_correct_eggs(self):
        correct_eggs = {
            'shproj': ["./egg1.egg", "./egg2.egg"],
            'notmeproj': ["./egg1.egg", "./egg2.egg"],
            'advanced_prod': ["./egg1.egg", "./egg2.egg"],
            'advanced_dev': ["./egg1.egg", "./egg2.egg"],
            'some': ["./egg3.egg", "./egg4.egg"]
        }
        for name, eggs in correct_eggs.items():
            target = self.conf.get_target_conf(name)
            self.assertEqual(target.eggs, eggs)

    def test_get_undefined(self):
        self.assertEqual(
            self.conf.get_target_conf('99'),
            _target(99, apikey='key'),
        )
        self.assertEqual(
            self.conf.get_target_conf('external/99', auth_required=False),
            _target(99, 'ext_endpoint', None),
        )
        with self.assertRaises(NotFoundException):
            self.conf.get_target_conf('nonexisting_ep/33')

    def test_get_invalid(self):
        # Missing target and no default defined
        with self.assertRaises(BadParameterException):
            self.conf.get_target_conf('default')
        # Bad project ID on command line
        with self.assertRaises(BadParameterException):
            self.conf.get_target_conf('99a')
        # Bad project ID in scrapinghub.yml
        conf = self._get_conf_with_yml(
            """
            projects:
                invalid: 50a
                invalid2: 123/external
            """)
        with self.assertRaises(BadConfigException):
            conf.get_target_conf('invalid')
        with self.assertRaises(BadConfigException):
            conf.get_target_conf('invalid2')

    def test_apikey_always_str(self):
        # API keys should always be strings, even if they contain only digits
        self.conf.apikeys['default'] = 123
        self.assertEqual(
            self.conf.get_target_conf('shproj'),
            _target(123, apikey='123'),
        )

    def test_hidden_apikey(self):
        self.conf.apikeys['default'] = 123
        self.assertEqual(
            self.conf.get_target_conf('shproj').apikey.__repr__(),
            "123"
        )
        self.conf.apikeys['default'] = "apikey_string"
        self.assertEqual(
            self.conf.get_target_conf('shproj').apikey.__repr__(),
            "apikeyXXXXXXX"
        )

    @mock.patch('shub.config.pwd_version', return_value='ver_AUTO')
    @mock.patch('shub.config.pwd_git_version', return_value='ver_GIT')
    @mock.patch('shub.config.pwd_hg_version', return_value='ver_HG')
    def test_get_version(self, mock_hg, mock_git, mock_ver):
        def _assert_version(version, expected):
            self.conf.version = version
            self.assertEqual(self.conf.get_version(), expected)
        _assert_version('GIT', 'ver_GIT')
        _assert_version('HG', 'ver_HG')
        _assert_version('somestring', 'somestring')
        _assert_version('', 'ver_AUTO')
        _assert_version('AUTO', 'ver_AUTO')


LOCAL_SCRAPINGHUB_YML = """
    projects:
        localextproj: external/123
    endpoints:
        external: local_ext_endpoint
    apikeys:
        external: key_ext
"""

GLOBAL_SCRAPY_CFG = textwrap.dedent("""
    [deploy]
    url = dotsc_endpoint
    username = dotsc_key

    [deploy:ext2]
    url = ext2_endpoint
    project = 333
    username = ext2_key
""")

NETRC = 'machine scrapinghub.com login netrc_key password ""'


class LoadShubConfigTest(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.globalpath = os.path.join(self.tmpdir, '.scrapinghub.yml')
        self.localpath = os.path.join(self.tmpdir, 'scrapinghub.yml')
        self.globalscrapycfgpath = os.path.join(self.tmpdir, '.scrapy.cfg')
        self.localscrapycfgpath = os.path.join(self.tmpdir, 'scrapy.cfg')
        self.netrcpath = os.path.join(self.tmpdir, '.netrc')
        with open(self.globalpath, 'w') as f:
            f.write(VALID_YAML_CFG)
        with open(self.localpath, 'w') as f:
            f.write(LOCAL_SCRAPINGHUB_YML)
        with open(self.globalscrapycfgpath, 'w') as f:
            f.write(GLOBAL_SCRAPY_CFG)
        with open(self.netrcpath, 'w') as f:
            f.write(NETRC)
        self._old_dir = os.getcwd()
        os.chdir(self.tmpdir)

        patcher_gsyp = mock.patch('shub.config.GLOBAL_SCRAPINGHUB_YML_PATH',
                                  new=self.globalpath)
        patcher_nrcp = mock.patch('shub.config.NETRC_PATH', new=self.netrcpath)
        patcher_gs = mock.patch('shub.config.get_sources',
                                return_value=[self.globalscrapycfgpath])
        self.addCleanup(patcher_gsyp.stop)
        self.addCleanup(patcher_nrcp.stop)
        self.addCleanup(patcher_gs.stop)
        patcher_gsyp.start()
        patcher_nrcp.start()
        patcher_gs.start()

    def tearDown(self):
        os.chdir(self._old_dir)
        shutil.rmtree(self.tmpdir)

    def test_scrapinghub_ymls_read(self):
        conf = load_shub_config()
        self.assertEqual(conf.get_apikey('shproj'), 'key')
        self.assertEqual(
            conf.get_endpoint('externalproj'),
            'local_ext_endpoint',
        )
        self.assertEqual(conf.get_apikey('externalproj'), 'key_ext')
        with self.assertRaises(BadParameterException):
            conf.get_project_id('ext2')

    def test_local_scrapinghub_yml_in_parent_dir(self):
        subsubdir = os.path.join(self.tmpdir, 'sub/sub')
        os.makedirs(subsubdir)
        os.chdir(subsubdir)
        conf = load_shub_config()
        self.assertEqual(conf.get_apikey('externalproj'), 'key_ext')

    def test_no_local_scrapinghub_yml(self):
        os.remove(self.localpath)
        conf = load_shub_config()
        self.assertEqual(conf.get_apikey('shproj'), 'key')
        with self.assertRaises(BadParameterException):
            conf.get_apikey('localextproj')

    def test_no_global_scrapinghub_yml(self):
        os.remove(self.globalpath)
        conf = load_shub_config()
        with self.assertRaises(BadParameterException):
            conf.get_apikey('shproj')
        self.assertEqual(conf.get_apikey('localextproj'), 'key_ext')

    def test_envvar_precedence(self):
        _old_environ = dict(os.environ)
        os.environ['SHUB_APIKEY'] = 'key_env'
        conf = load_shub_config()
        self.assertEqual(conf.get_apikey('shproj'), 'key_env')
        os.environ.clear()
        os.environ.update(_old_environ)

    def test_autocreate_empty_global_scrapinghub_yml(self):
        os.remove(self.globalpath)
        os.remove(self.globalscrapycfgpath)
        os.remove(self.netrcpath)
        load_shub_config()
        self.assertTrue(os.path.isfile(self.globalpath))
        with open(self.globalpath) as f:
            self.assertEqual(f.read(), "")

    def test_automigrate_to_global_scrapinghub_yml(self):
        def _check_conf():
            conf = load_shub_config()
            self.assertEqual(
                conf.get_target('123'),
                (123, 'dotsc_endpoint', 'netrc_key'),
            )
            self.assertEqual(conf.projects['ext2'], 'ext2/333')
            self.assertEqual(
                conf.get_target('ext2'),
                (333, 'ext2_endpoint', 'ext2_key'),
            )
        os.remove(self.globalpath)
        _check_conf()
        self.assertTrue(os.path.isfile(self.globalpath))
        os.remove(self.netrcpath)
        os.remove('.scrapy.cfg')
        _check_conf()

    def test_automigrate_project_scrapy_cfg(self):
        def _check_conf():
            conf = load_shub_config()
            self.assertEqual(
                conf.get_target('default'),
                (222, 'scrapycfg_endpoint/', 'key'),
            )
            self.assertEqual(
                conf.get_target('ext2'),
                (333, 'ext2_endpoint/', 'ext2_key'),
            )
            self.assertEqual(
                conf.get_target('ext3'),
                (333, 'scrapycfg_endpoint/', 'key'),
            )
            self.assertEqual(
                conf.get_target('ext4'),
                (444, 'scrapycfg_endpoint/', 'ext4_key'),
            )
            self.assertEqual(conf.get_version(), 'ext2_ver')
        scrapycfg = """
            [deploy]
            project = 222
            url = scrapycfg_endpoint/scrapyd/

            [deploy:ext2]
            url = ext2_endpoint/scrapyd/
            project = 333
            username = ext2_key
            version = ext2_ver

            [deploy:ext3]
            project = 333

            [deploy:ext4]
            project = 444
            username = ext4_key
        """
        with open(self.localscrapycfgpath, 'w') as f:
            f.write(textwrap.dedent(scrapycfg))
        os.mkdir('project')
        os.chdir('project')
        conf = load_shub_config()
        with self.assertRaises(BadParameterException):
            conf.get_target('ext2')
        os.remove(self.localpath)
        # Loaded from scrapy.cfg
        _check_conf()
        # Same config should now be loaded from scrapinghub.yml
        self.assertTrue(os.path.isfile(self.localpath))
        _check_conf()


class ConfigHelpersTest(unittest.TestCase):

    @mock.patch('shub.config.load_shub_config')
    def test_get_target_version(self, mock_lsh):
        get_target('mytarget', auth_required=False)
        get_target_conf('mytargetconf', auth_required=False)
        get_version()
        mock_lsh.return_value.get_target.assert_called_once_with(
            'mytarget', auth_required=False)
        mock_lsh.return_value.get_target_conf.assert_called_once_with(
            'mytargetconf', auth_required=False)
        mock_lsh.return_value.get_version.assert_called_once_with()



import json
import stat
import shlex
import signal
import os.path
from os import stat as os_stat
from shutil import copyfile

import click

from shub.config import load_shub_config
from shub.image import utils


SHORT_HELP = 'Run custom image locally.'
HELP = """
Run a custom Docker image locally.

The command should be helpful to ensure that your custom image is properly
written and do some preliminary local tests before pushing it to Scrapy Cloud.

Most of the command parameters coincide with parameters for 'shub schedule'
command to simplfy its usage.

The `spider` argument should match the spider's name, e.g.:

    shub image run myspider

A more advanced example of using non-default target with settings/arguments:

    shub image run production/myspider -a ARG1=VAL1 -s LOG_LEVEL=DEBUG
"""

SCRAPINGHUB_VOLUME = '/scrapinghub'
WRAPPER_FILENAME = 'start-crawl-local'
WRAPPER_LOCAL_PATH = os.path.join(os.path.dirname(__file__), 'wrapper.py')
WRAPPER_IMAGE_PATH = os.path.join(SCRAPINGHUB_VOLUME, WRAPPER_FILENAME)


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("spider", type=click.STRING)
@click.option('-a', '--argument', 'args',
              help='Spider argument (-a name=value)', multiple=True)
@click.option('-s', '--set', 'settings',
              help='Job-specific setting (-s name=value)', multiple=True)
@click.option('-e', '--environment', multiple=True,
              help='Job environment variable (-e VAR=VAL)')
@click.option("-V", "--version", help="use custom release version")
@click.option("-v", "--verbose", is_flag=True,
              help="stream additional logs to console")
@click.option("-k", "--keep-volume", help="Keep volume folder", is_flag=True)
def cli(spider, args, settings, environment, version, verbose, keep_volume):
    run_cmd(spider, args, settings, environment, version, keep_volume)


def run_cmd(spider, args, settings, environment, version, keep_volume):
    try:
        target, spider = spider.rsplit('/', 1)
    except ValueError:
        target = 'default'

    config = load_shub_config()
    image = config.get_image(target)
    version = version or config.get_version()
    image_name = utils.format_image_name(image, version)
    docker_client = utils.get_docker_client()

    env = _format_environment(spider, args, settings, environment)
    _run_with_docker(docker_client, image_name, env, keep_volume)


def _format_environment(spider, args, settings, environment):
    """Convert all input crawl args to environment variables."""
    # required defaults, can be overwritten with meta if needed
    job_data = {'spider': spider, 'key': '1/2/3', 'auth': '<auth>'}

    args = dict(x.split('=', 1) for x in args)
    cmd_args = shlex.split(args.pop('cmd_args', ''))
    if spider.startswith('py:'):
        job_data['job_cmd'] = [spider] + cmd_args
    else:
        job_data['spider_args'] = args
    meta = args.pop('meta', None)
    if meta:
        job_data.update(json.loads(meta))

    job_environment = dict(x.split('=', 1) for x in environment)
    job_settings = dict(x.split('=', 1) for x in settings)
    return {
        'SHUB_JOBKEY': job_data['key'],
        'SHUB_SPIDER': spider,
        'SHUB_JOB_DATA': _json_dumps(job_data),
        'SHUB_JOB_ENV': _json_dumps(job_environment),
        'SHUB_SETTINGS': _json_dumps({'job_settings': job_settings}),
        'PYTHONUNBUFFERED': 1,
    }


def _json_dumps(data):
    return json.dumps(data, sort_keys=True, separators=(',', ':'))


def _run_with_docker(client, image_name, env, keep_volume=False):
    """Run a local docker container with the given custom image."""

    def _signal_handler(sig, _):
        client.kill(container, sig)

    tmpdir_kw = {'prefix': 'shub-image-run-', 'cleanup': not keep_volume}
    with utils.make_temp_directory(**tmpdir_kw) as volume_dir:
        container = _create_container(client, image_name, env, volume_dir)
        try:
            client.start(container)
            signal.signal(signal.SIGINT, _signal_handler)
            signal.signal(signal.SIGTERM, _signal_handler)
            for log in client.logs(container, stream=True):
                click.echo(log.rstrip())
        finally:
            client.remove_container(container, force=True)


def _create_container(client, image_name, environment, volume_dir):
    """Create a docker container and customize its setup."""
    # copy start-crawl wrapper to the volume temporary directory
    wrapper_cont_path = os.path.join(volume_dir, WRAPPER_FILENAME)
    copyfile(WRAPPER_LOCAL_PATH, wrapper_cont_path)
    wrapper_perms = os_stat(wrapper_cont_path).st_mode | stat.S_IEXEC
    os.chmod(wrapper_cont_path, wrapper_perms)  # must be executable
    fifo_path = os.path.join(volume_dir, 'scrapinghub.fifo')
    environment['SHUB_FIFO_PATH'] = fifo_path
    # keep using default /scrapinghub volume but mount it as a temporary
    # directory in the host /tmp/ to have access to the files in needed
    binds = {volume_dir: {'bind': SCRAPINGHUB_VOLUME, 'mode': 'rw'}}
    host_config = client.create_host_config(binds=binds)
    return client.create_container(
        image=image_name,
        command=[WRAPPER_IMAGE_PATH],
        environment=environment,
        volumes=[volume_dir],
        host_config=host_config,
    )



import click

from shub.config import list_targets_callback
from shub.image import build
from shub.image import push
from shub.image import deploy
from shub.image import utils


SHORT_HELP = "Shortcut command for build-push-deploy chain"
HELP = """
Upload command is a handy shortcut to rebuild and redeploy your project
(in other words it does consecutive calls of build-push-deploy cmds).

Obviously it accepts all the options for the commands above.
"""


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True,
              help="stream upload logs to console")
@click.option("-V", "--version", help="release version")
@click.option("--username", help="docker registry name")
@click.option("--password", help="docker registry password")
@click.option("--email", help="docker registry email")
@click.option("--apikey", help="SH apikey to use built-in registry")
@click.option("--insecure", is_flag=True, help="use insecure registry")
@click.option("--async", "async_", is_flag=True, help="[DEPRECATED] enable asynchronous mode",
              callback=utils.deprecate_async_parameter)
@click.option("-S", "--skip-tests", help="skip testing image", is_flag=True)
@click.option("-R", "--reauth", is_flag=True,
              help="re-authenticate to registry before pushing")
@click.option("-n", "--no-cache", is_flag=True,
              help="Do not use cache when building the image")
@click.option("-b", "--build-arg", multiple=True,
              help="Allow to pass build arguments to docker client.")
@click.option("-f", "--file", "filename", default='Dockerfile',
              help="Name of the Dockerfile (Default is 'PATH/Dockerfile')")
def cli(target, debug, verbose, version, username, password, email,
        apikey, insecure, async_, skip_tests, reauth, no_cache, build_arg, filename):
    upload_cmd(target, version, username, password, email, apikey, insecure,
               async_, skip_tests, reauth, no_cache, build_arg, filename)


def upload_cmd(target, version, username=None, password=None, email=None,
               apikey=None, insecure=False, async_=False, skip_tests=False,
               reauth=False, no_cache=False, build_arg=(), filename='Dockerfile'):
    build.build_cmd(target, version, skip_tests, no_cache, build_arg, filename=filename)
    # skip tests for push command anyway because they run in build command if not skipped
    push.push_cmd(target, version, username, password, email, apikey,
                  insecure, skip_tests=True, reauth=reauth)
    deploy.deploy_cmd(target, version, username, password, email,
                      apikey, insecure, async_)



import json
import os
import stat
import sys
import unittest
import textwrap
import time
from io import StringIO
from unittest.mock import Mock, MagicMock, patch

import click
import yaml
from click.testing import CliRunner
from collections import deque
from scrapinghub import ScrapinghubAPIError

from shub import utils
from shub.config import ShubConfig
from shub.exceptions import (
    BadParameterException, InvalidAuthException, MissingAuthException,
    NotFoundException, RemoteErrorException, SubcommandException
)

from .utils import AssertInvokeRaisesMixin, mock_conf


class UtilsTest(AssertInvokeRaisesMixin, unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    @patch('shub.utils.sys.frozen', new=True, create=True)
    @patch('shub.utils.find_exe', return_value='/my/python')
    def test_patch_sys_executable(self, mock_find_exe):
        original_exe = sys.executable
        with patch('shub.utils.sys.frozen', new=False):
            with utils.patch_sys_executable():
                self.assertEqual(sys.executable, original_exe)
        with utils.patch_sys_executable():
            self.assertEqual(sys.executable, '/my/python')
        # Make sure we properly cleaned up after ourselves
        self.assertEqual(sys.executable, original_exe)
        mock_find_exe.side_effect = NotFoundException
        with self.assertRaises(NotFoundException):
            with utils.patch_sys_executable():
                pass

    @patch('shub.utils.which')
    def test_find_exe(self, mock_fe):
        mock_fe.return_value = '/usr/bin/python'
        self.assertEqual(utils.find_exe('python'), '/usr/bin/python')
        mock_fe.return_value = None
        with self.assertRaises(NotFoundException):
            utils.find_exe('python')

    def test_run_cmd_captures_stderr(self):
        cmd = [
            'python', '-c',
            # The next two lines are ONE list element (no comma)
            'from __future__ import print_function; import sys; '
            'print("Hello", file=sys.stderr)',
        ]
        self.assertEqual(utils.run_cmd(cmd), '')
        with self.assertRaisesRegex(SubcommandException, r'STDERR[\s-]+Hello'):
            cmd[-1] += '; sys.exit(99)'
            utils.run_cmd(cmd)

    def test_pwd_git_version_without_git(self):
        # Change into test dir to make sure we're within a repo
        os.chdir(os.path.dirname(__file__))
        self.assertIsNotNone(utils.pwd_git_version())
        with patch('shub.utils.which', return_value=None):
            self.assertIsNone(utils.pwd_git_version())

    @patch('shub.utils.pwd_git_version', return_value='ver_GIT')
    @patch('shub.utils.pwd_hg_version', return_value='ver_HG')
    @patch('shub.utils.pwd_bzr_version', return_value='ver_BZR')
    @patch('shub.utils.time.time', return_value=101)
    def test_pwd_version(self, mock_time, mock_bzr, mock_hg, mock_git):
        self.assertEqual(utils.pwd_version(), 'ver_GIT')
        mock_git.return_value = None
        self.assertEqual(utils.pwd_version(), 'ver_HG')
        mock_hg.return_value = None
        self.assertEqual(utils.pwd_version(), 'ver_BZR')
        mock_bzr.return_value = None
        with self.runner.isolated_filesystem():
            with open('setup.py', 'w') as f:
                f.write("from setuptools import setup\n")
                f.write("setup(version='1.0')")
            self.assertEqual(utils.pwd_version(), '1.0')
            setup_version = (
                'Building lxml version 3.4.4.'
                '\nBuilding without Cython.'
                '\nUsing build configuration of libxslt 1.1.28'
                '\n3.4.4'
            )
            with patch('shub.utils.run_python', return_value=setup_version):
                self.assertEqual(utils.pwd_version(), '3.4.4')
            os.mkdir('subdir')
            os.chdir('subdir')
            self.assertEqual(utils.pwd_version(), '101')
            open('../scrapy.cfg', 'w').close()
            self.assertEqual(utils.pwd_version(), '1.0')

    @patch('shub.utils.pwd_git_version')
    def test_pwd_version_clean(self, mock_git):
        mock_git.return_value = 'vers_1'
        self.assertEqual(utils.pwd_version(), 'vers_1')
        mock_git.return_value = 've  rs _ 2'
        self.assertEqual(utils.pwd_version(), 'vers_2')
        mock_git.return_value = 'vers -3_1:1'
        self.assertEqual(utils.pwd_version(), 'vers-3_11')
        mock_git.return_value = 'vers -4_1!$@%#&$()2'
        self.assertEqual(utils.pwd_version(), 'vers-4_12')

    def test_get_job_specs(self):
        conf = mock_conf(self)

        def _test_specs(job, expected_job_id, expected_endpoint):
            self.assertEqual(
                utils.get_job_specs(job),
                (expected_job_id, conf.get_apikey(expected_endpoint)),
            )
        _test_specs('10/20/30', '10/20/30', 'default')
        _test_specs('2/3', '1/2/3', 'default')
        _test_specs('default/2/3', '1/2/3', 'default')
        _test_specs('prod/2/3', '2/2/3', 'default')
        _test_specs('vagrant/2/3', '3/2/3', 'vagrant')
        _test_specs(
            'https://app.zyte.com/p/7389/259/1/#/log/line/0',
            '7389/259/1',
            'default',
        )
        _test_specs(
            'https://app.zyte.com/p/7389/job/259/1/',
            '7389/259/1',
            'default',
        )

    def test_get_job_specs_validates_jobid(self):
        invalid_job_ids = ['/1/1', '123', '1/2/a', '1//']
        for job_id in invalid_job_ids:
            with self.assertRaises(BadParameterException):
                utils.get_job_specs(job_id)

    @patch('shub.utils.HubstorageClient', autospec=True)
    def test_get_job(self, mock_HSC):
        class MockJob:
            metadata = {'some': 'val'}
        mockjob = MockJob()
        mock_HSC.return_value.get_job.return_value = mockjob
        conf = mock_conf(self)

        self.assertIs(utils.get_job('1/1/1'), mockjob)
        mock_HSC.assert_called_once_with(auth=conf.apikeys['default'])

        with self.assertRaises(BadParameterException):
            utils.get_job('1/1/')

        # Non-existent job
        mockjob.metadata = None
        with self.assertRaises(NotFoundException):
            utils.get_job('1/1/1')

    def test_is_deploy_successful(self):
        # no results
        last_logs = deque(maxlen=5)
        assert not utils._is_deploy_successful(last_logs)
        # missing or incorrect data
        last_logs.append("")
        assert not utils._is_deploy_successful(last_logs)
        last_logs.append("abcdef")
        assert not utils._is_deploy_successful(last_logs)
        last_logs.append('{"field":"wrong"}')
        assert not utils._is_deploy_successful(last_logs)
        # error status
        last_logs.append('{"status":"error"}')
        assert not utils._is_deploy_successful(last_logs)
        # successful status
        last_logs.append('{"status":"ok"}')
        assert utils._is_deploy_successful(last_logs)
        last_logs.append('{"field":"value","status":"ok"}')
        assert utils._is_deploy_successful(last_logs)
        # more complex python expression
        last_logs.append('{"status":"ok", "project": 1111112, '
                         '"version": "1234-master", "spiders": 3}')
        assert utils._is_deploy_successful(last_logs)

    def test_job_live(self):
        job = MagicMock()
        job._metadata_updated = time.time()
        for live_value in ('pending', 'running'):
            job.metadata.__getitem__.return_value = live_value
            self.assertTrue(utils.job_live(job))
        for dead_value in ('finished', 'deleted'):
            job.metadata.__getitem__.return_value = dead_value
            self.assertFalse(utils.job_live(job))

    def test_job_live_updates_metadata(self):
        job = MagicMock(spec=['metadata'])
        with patch('shub.utils.time.time') as mock_time:
            mock_time.return_value = 0
            utils.job_live(job)
            mock_time.return_value = 10
            utils.job_live(job, refresh_meta_after=20)
            self.assertFalse(job.metadata.expire.called)
            utils.job_live(job, refresh_meta_after=5)
            self.assertTrue(job.metadata.expire.called)
            job.metadata.expire.reset_mock()
            utils.job_live(job, refresh_meta_after=5)
            self.assertFalse(job.metadata.expire.called)

    @patch('shub.utils.time.sleep')
    def test_job_resource_iter(self, mock_sleep):
        job = MagicMock(spec=['key', 'metadata', 'resource'])
        job.key = 'jobkey'
        job.metadata = {'state': 'running'}

        def make_items(iterable):
            return [json.dumps({'_key': x}) for x in iterable]

        def magic_iter(*args, **kwargs):
            """
            Return two different iterators on the first two calls, set job's
            state to 'finished' after the second call.
            """
            if magic_iter.stage == 0:
                if 'startafter' in kwargs:
                    self.assertEqual(kwargs['startafter'], None)
                magic_iter.stage = 1
                return iter(make_items([1, 2, 3]))
            elif magic_iter.stage == 1:
                self.assertEqual(kwargs['startafter'], 3)
                magic_iter.stage = 0
                job.metadata = {'state': 'finished'}
                return iter(make_items([4, 5, 6]))
            elif magic_iter.stage == 2:
                self.assertEqual(kwargs['startafter'], 'jobkey/996')
                return iter([])

        def jri_result(follow, tail=None):
            return list(utils.job_resource_iter(
                job,
                job.resource,
                follow=follow,
                tail=tail,
                output_json=True,
            ))

        job.resource.iter_json = magic_iter

        magic_iter.stage = 0
        self.assertEqual(jri_result(False), make_items([1, 2, 3]))
        self.assertFalse(mock_sleep.called)

        magic_iter.stage = 0
        self.assertEqual(jri_result(True), make_items([1, 2, 3, 4, 5, 6]))
        self.assertTrue(mock_sleep.called)

        magic_iter.stage = 0
        job.metadata = {'state': 'finished'}
        self.assertEqual(jri_result(True), make_items([1, 2, 3]))

        magic_iter.stage = 2
        job.resource.stats.return_value = {'totals': {'input_values': 1000}}
        self.assertEqual(jri_result(True, tail=3), [])

    @patch('shub.utils.requests.get', autospec=True)
    def test_latest_github_release(self, mock_get):
        with self.runner.isolated_filesystem():
            mock_get.return_value.json.return_value = {'key': 'value'}
            self.assertLessEqual(
                {'key': 'value'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            mock_get.return_value.json.return_value = {'key': 'newvalue'}
            self.assertLessEqual(
                {'key': 'value'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            self.assertLessEqual(
                {'key': 'newvalue'}.items(),
                utils.latest_github_release(force_update=True,
                                            cache='./cache.txt').items(),
            )
            # Garbage in cache
            mock_get.return_value.json.return_value = {'key': 'value'}
            with open('./cache.txt', 'w') as f:
                f.write('abc')
            self.assertLessEqual(
                {'key': 'value'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            mock_get.return_value.json.return_value = {'key': 'newvalue'}
            self.assertLessEqual(
                {'key': 'value'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            # Readonly cache file
            mock_get.return_value.json.return_value = {'key': 'value'}
            with open('./cache.txt', 'w') as f:
                f.write('abc')
            os.chmod('./cache.txt', stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
            self.assertLessEqual(
                {'key': 'value'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            mock_get.return_value.json.return_value = {'key': 'newvalue'}
            self.assertLessEqual(
                {'key': 'newvalue'}.items(),
                utils.latest_github_release(cache='./cache.txt').items(),
            )
            with open('./cache.txt') as f:
                self.assertEqual(f.read(), 'abc')

    @patch('shub.utils.latest_github_release', autospec=True)
    @patch('shub.utils.shub.__version__', new='1.5.0')
    def test_update_available(self, mock_lgr):
        class MockException(Exception):
            pass
        mock_lgr.return_value = {'name': 'v1.4.0', 'html_url': 'link'}
        self.assertIsNone(utils.update_available())
        mock_lgr.return_value['name'] = 'v1.5.0'
        self.assertIsNone(utils.update_available())
        mock_lgr.return_value['name'] = 'v1.5.1'
        self.assertEqual(utils.update_available(), 'link')
        mock_lgr.return_value['name'] = 'v1.6.1'
        self.assertEqual(utils.update_available(), 'link')
        mock_lgr.return_value['name'] = 'v2.0.0'
        self.assertEqual(utils.update_available(), 'link')
        mock_lgr.return_value = {'error': 'unavailable'}
        self.assertIsNone(utils.update_available())
        mock_lgr.return_value = None
        self.assertIsNone(utils.update_available())
        mock_lgr.side_effect = MockException
        self.assertIsNone(utils.update_available())
        with self.assertRaises(MockException):
            utils.update_available(silent_fail=False)

    @patch('shub.utils.pip_main', autospec=True)
    @patch('shub.utils.pip', autospec=True)
    def test_download_from_pypi(self, mock_pip, mock_pip_main):
        def _call(*args, **kwargs):
            utils.download_from_pypi(*args, **kwargs)
            return mock_pip_main.call_args[0][0]

        with self.assertRaises(ValueError):
            utils.download_from_pypi('tmpdir')
        with self.assertRaises(ValueError):
            utils.download_from_pypi('tmpdir', pkg='shub', reqfile='req.txt')
        self.assertFalse(mock_pip_main.called)

        # 1.0 (Ubuntu Precise)
        del mock_pip.__version__
        pipargs = _call('tmpdir', pkg='shub')
        self.assertNotIn('--no-use-wheel', pipargs)

        # 1.5.4 (Ubuntu Trusty)
        mock_pip.__version__ = '1.5.4'
        pipargs = _call('tmpdir', pkg='shub', extra_args=['--x'])
        self.assertEqual(pipargs[0], 'install')
        for expected_arg in ['--no-use-wheel', '--no-deps', 'shub', '--x']:
            self.assertIn(expected_arg, pipargs)
        # Make sure list contains '-d' followed by 'tmpdir'
        self.assertEqual(pipargs.index('-d') + 1, pipargs.index('tmpdir'))
        pipargs = _call('tmpdir', reqfile='req.txt')
        self.assertEqual(pipargs.index('-r') + 1, pipargs.index('req.txt'))

        # Replace deprecated commands in newer versions
        mock_pip.__version__ = '7.1.2.dev0'
        pipargs = _call('tmpdir', pkg='shub')
        self.assertEqual(pipargs[0], 'install')
        self.assertIn('--no-binary=:all:', pipargs)
        mock_pip.__version__ = '8.0.2'
        pipargs = _call('tmpdir', pkg='shub')
        self.assertEqual(pipargs[0], 'download')
        self.assertIn('--no-binary=:all:', pipargs)

    def test_echo_short_log_if_deployed(self):
        last_logs = ["last log line"]

        deployed = True
        for verbose, expected in ((True, ""), (False, "last log line\n")):
            with patch('sys.stdout', new_callable=StringIO) as stdout:
                utils.echo_short_log_if_deployed(deployed, last_logs, verbose=verbose)
            self.assertEqual(expected, stdout.getvalue())

        deployed = False
        for verbose, expected in ((True, ""), (False, "Deploy log last 1 lines:\nlast log line\n")):
            with patch('sys.stdout', new_callable=StringIO) as stdout:
                utils.echo_short_log_if_deployed(deployed, last_logs, verbose=verbose)
            self.assertEqual(expected, stdout.getvalue())

    def test_write_and_echo_logs(self):
        last_logs = []
        rsp = Mock()
        rsp.iter_lines = Mock(return_value=iter([b"line1", b"line2"]))
        self.assertRaises(RemoteErrorException,
                          utils.write_and_echo_logs,
                          keep_log=True, last_logs=last_logs,
                          rsp=rsp, verbose=True)
        self.assertEqual(last_logs, [b"line1", b"line2"])
        last_logs = []

        rsp.iter_lines = Mock(return_value=iter(
            [b"line1", b'{"status":"ok","fieldK":"fieldV"}']))
        utils.write_and_echo_logs(keep_log=True, last_logs=last_logs,
                                  rsp=rsp, verbose=True)
        self.assertEqual(last_logs, [
            b"line1", b'{"status":"ok","fieldK":"fieldV"}'])

    def test_update_yaml_dict(self):
        YAML_BEFORE = textwrap.dedent("""\
            a:
              unrelated: dict
            b:
              key1: val1
              key2: val2
        """)
        DICT_EXPECTED = {
            'a': {'unrelated': 'dict'},
            'b': {'key1': 'newval1', 'key2': 'val2', 'key3': 'val3'}
        }
        runner = CliRunner()
        with runner.isolated_filesystem():
            with open('conf.yml', 'w') as f:
                f.write(YAML_BEFORE)
            with utils.update_yaml_dict('conf.yml') as conf:
                conf['b']['key1'] = 'newval1'
                conf['b']['key3'] = 'val3'
            with open('conf.yml') as f:
                self.assertEqual(yaml.safe_load(f), DICT_EXPECTED)
                f.seek(0)
                self.assertIn("key1: newval1", f.read())

    def test_update_yaml_dict_handles_file_errors(self):
        with CliRunner().isolated_filesystem():
            self.assertFalse(os.path.isfile('didnt_exist.yml'))
            with utils.update_yaml_dict('didnt_exist.yml'):
                pass
            self.assertTrue(os.path.isfile('didnt_exist.yml'))

            os.mkdir('a_directory')
            with self.assertRaises(IOError):
                with utils.update_yaml_dict('a_directory'):
                    pass

    @patch('shub.config.GLOBAL_SCRAPINGHUB_YML_PATH', 'global.yml')
    def test_update_yaml_dict_uses_global_by_default(self):
        @click.command()
        def call_update_yaml_dict():
            with utils.update_yaml_dict():
                pass

        runner = CliRunner()
        with runner.isolated_filesystem():
            result = runner.invoke(call_update_yaml_dict)
        assert 'deprecated' in result.output

    @patch('shub.utils.ScrapinghubClient')
    def test_has_project_access(self, mock_client):
        mock_client.return_value.projects.list.side_effect = ScrapinghubAPIError(
            'Authentication failed')
        with self.assertRaises(InvalidAuthException):
            utils.has_project_access(12345, 'mock_endpoint', 'abcdef')
        mock_client.return_value.projects.list.side_effect = ScrapinghubAPIError(
            'Random error')
        with self.assertRaises(RemoteErrorException):
            utils.has_project_access(12345, 'mock_endpoint', 'abcdef')

    def test_get_project_dir(self):
        # OSX test work-around: /var/ is symlinked to /private/var/,
        # and tempfile.mkdtemp() returns a non-absolute path while
        # os.path.dirname function is resolving the absolute location.
        # The workaround patches a path returned by utils.get_project_dir
        def get_project_dir():
            project_dir = utils.get_project_dir()
            if project_dir.startswith('/private/'):
                project_dir = project_dir[8:]
            return project_dir

        with CliRunner().isolated_filesystem() as basepath:
            os.makedirs('a/b/c')
            os.chdir('a/b/c')
            with self.assertRaises(NotFoundException):
                get_project_dir()
            open(os.path.join(basepath, 'Dockerfile'), 'w').close()
            self.assertEqual(get_project_dir(), basepath)
            # scrapy.cfg takes precedence over Dockerfile
            open(os.path.join(basepath, 'a', 'scrapy.cfg'), 'w').close()
            self.assertEqual(
                get_project_dir(),
                os.path.join(basepath, 'a'))
            # scrapinghub.yml takes precedence over both
            open(os.path.join(basepath, 'a', 'b', 'scrapinghub.yml'), 'w'
                 ).close()
            self.assertEqual(
                get_project_dir(),
                os.path.join(basepath, 'a', 'b'))


class OnboardingWizardTestCase(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.has_project_access = True

    @patch('shub.utils.has_project_access')
    def _test_wizard(self, mock_project_access, conf=None, target='default',
                     image=None, sh_yml=None, scrapy_cfg=True,
                     dockerfile=False, **kwargs):
        if not conf:
            conf = ShubConfig()
            conf.apikeys = {'default': 'abcdef'}
        mock_project_access.return_value = self.has_project_access

        @click.command()
        def call_wizard():
            utils.create_scrapinghub_yml_wizard(
                conf, target=target, image=image)

        with self.runner.isolated_filesystem():
            if scrapy_cfg:
                open('scrapy.cfg', 'w').close()
            if dockerfile:
                open('Dockerfile', 'w').close()
            if sh_yml:
                with open('scrapinghub.yml', 'w') as f:
                    f.write(sh_yml)
                conf.load_file('scrapinghub.yml')
            result = self.runner.invoke(call_wizard, **kwargs)
            if os.path.exists('scrapinghub.yml'):
                with open('scrapinghub.yml') as f:
                    sh_yml = yaml.safe_load(f.read())
            else:
                sh_yml = None
            return result, conf, sh_yml

    def test_outside_of_project(self):
        result, _, sh_yml = self._test_wizard(scrapy_cfg=False)
        assert result.exit_code == NotFoundException.exit_code
        assert sh_yml is None

    def test_not_logged_in(self):
        conf = ShubConfig()
        result, _, sh_yml = self._test_wizard(conf=conf)
        assert result.exit_code == MissingAuthException.exit_code
        assert sh_yml is None

    def test_scrapy_project(self):
        result, conf, sh_yml = self._test_wizard(input='12345\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert sh_yml == {'project': 12345}

    def test_custom_project(self):
        result, conf, sh_yml = self._test_wizard(
            scrapy_cfg=False, dockerfile=True, input='12345\n\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert conf.images == {'default': True}
        assert sh_yml == {'project': 12345, 'image': True}

    def test_custom_repository(self):
        result, conf, sh_yml = self._test_wizard(
            scrapy_cfg=False, dockerfile=True, input='12345\nrepo\n')
        assert result.exit_code == 0
        assert conf.images == {'default': 'repo'}
        assert sh_yml == {'project': 12345, 'image': 'repo'}

    def test_ambiguous_project_not_custom(self):
        result, conf, sh_yml = self._test_wizard(
            scrapy_cfg=True, dockerfile=True, input='12345\nn\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert not conf.images
        assert sh_yml == {'project': 12345}

    def test_ambiguous_project_custom(self):
        result, conf, sh_yml = self._test_wizard(
            scrapy_cfg=True, dockerfile=True, input='12345\n\n\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert conf.images == {'default': True}
        assert sh_yml == {'project': 12345, 'image': True}

    def test_never_ask_for_repository_when_image_false(self):
        result, conf, _ = self._test_wizard(
            image=False, scrapy_cfg=False, dockerfile=True, input='12345\n')
        assert result.exit_code == 0
        assert 'repository' not in result.output
        assert not conf.images

    def test_always_ask_for_repository_when_image_true(self):
        result, conf, sh_yml = self._test_wizard(
            image=True, input='12345\nrepo\n')
        assert result.exit_code == 0
        assert conf.images == {'default': 'repo'}
        assert sh_yml == {'project': 12345, 'image': 'repo'}

    def test_skip_on_existing_scrapinghub_yml_and_not_image(self):
        original_sh_yml = 'project: 12345\n'
        result, conf, sh_yml = self._test_wizard(sh_yml=original_sh_yml)
        assert result.exit_code == 0
        assert not result.output
        assert not conf.images
        assert sh_yml == {'project': 12345}

    def test_add_image_for_existing_default_target(self):
        original_sh_yml = 'projects:\n  default: 12345\n  prod: 33333\n'
        result, conf, sh_yml = self._test_wizard(
            image=True, sh_yml=original_sh_yml, input='\n')
        assert result.exit_code == 0
        assert conf.images == {'default': True}
        assert sh_yml == {
            'projects': {'default': 12345, 'prod': 33333},
            'image': True
        }

    def test_add_image_for_existing_nondefault_target(self):
        original_sh_yml = 'projects:\n  default: 12345\n  prod: 33333\n'
        result, conf, sh_yml = self._test_wizard(
            target='prod', image=True, sh_yml=original_sh_yml, input='\n')
        assert result.exit_code == 0
        assert not conf.images
        assert sh_yml == {
            'projects': {
                'default': 12345,
                'prod': {
                    'id': 33333,
                    'image': True,
                },
            },
        }

    def test_skip_for_already_defined_default_image(self):
        original_sh_yml = 'project: 12345\nimage: true\n'
        result, conf, sh_yml = self._test_wizard(
            image=True, sh_yml=original_sh_yml)
        assert result.exit_code == 0
        assert not result.output

    def test_skip_for_already_defined_nondefault_image(self):
        original_sh_yml = textwrap.dedent("""\
            projects:
              default: 12345
              prod:
                id: 33333
                image: true\
            """)
        result, conf, sh_yml = self._test_wizard(
            target='prod', image=True, sh_yml=original_sh_yml)
        assert result.exit_code == 0
        assert not result.output

    def test_scrapy_project_with_numeric_target(self):
        result, conf, sh_yml = self._test_wizard(target='12345')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert sh_yml == {'project': 12345}

    def test_custom_project_with_numeric_target(self):
        result, conf, sh_yml = self._test_wizard(
            target='12345', scrapy_cfg=False, dockerfile=True, input='repo\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345}
        assert conf.images == {'default': 'repo'}
        assert sh_yml == {'project': 12345, 'image': 'repo'}

    def test_dont_leak_global_config(self):
        conf = ShubConfig()
        conf.projects = {'prod': 33333}
        conf.apikeys = {'default': 'abc'}
        result, conf, sh_yml = self._test_wizard(conf=conf, input='12345\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345, 'prod': 33333}
        assert conf.apikeys == {'default': 'abc'}
        assert sh_yml == {'project': 12345}

    def test_dont_leak_global_config_on_image(self):
        conf = ShubConfig()
        conf.projects = {'prod': 33333}
        conf.apikeys = {'default': 'abc'}
        result, conf, sh_yml = self._test_wizard(
            conf=conf, scrapy_cfg=False, dockerfile=True,
            input='12345\nrepo\n')
        assert result.exit_code == 0
        assert conf.projects == {'default': 12345, 'prod': 33333}
        assert conf.apikeys == {'default': 'abc'}
        assert conf.images == {'default': 'repo'}
        assert sh_yml == {'project': 12345, 'image': 'repo'}



from collections import OrderedDict

import click

from shub import exceptions as shub_exceptions
from shub.config import load_shub_config, list_targets_callback
from shub.image import utils
from shub.image.test import test_cmd
from shub.image.utils import get_image_registry

SHORT_HELP = 'Push an image to a specified docker registry'

HELP = """
A command to push your image to specified docker registry.

The command is a simple wrapper for `docker push` command and uses docker
daemon on your system to build an image. The only differences are that it
can generate correct image version and provide easy registry login logic.

The optional params are mostly related with registry authorization.
By default, the tool tries to call the registry in insecure manner,
otherwise you have to enter your credentials (at least username/password).
"""

LOGIN_ERROR_MSG = 'Please authorize with docker login'


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True,
              help="stream push logs to console")
@click.option("-V", "--version", help="release version")
@click.option("--username", help="docker registry name")
@click.option("--password", help="docker registry password")
@click.option("--email", help="docker registry email")
@click.option("--apikey", help="SH apikey to use built-in registry")
@click.option("--insecure", is_flag=True, help="use insecure registry")
@click.option("-S", "--skip-tests", help="skip testing image", is_flag=True)
@click.option("-R", "--reauth", is_flag=True,
              help="re-authenticate to registry")
def cli(target, debug, verbose, version, username, password, email, apikey,
        insecure, skip_tests, reauth):
    push_cmd(target, version, username, password, email, apikey, insecure,
             skip_tests, reauth)


def push_cmd(target, version, username, password, email, apikey, insecure,
             skip_tests, reauth):
    # Test the image content after building it
    if not skip_tests:
        test_cmd(target, version)

    client = utils.get_docker_client()
    config = load_shub_config()
    image = config.get_image(target)
    username, password = utils.get_credentials(
        username=username, password=password, insecure=insecure,
        apikey=apikey, target_apikey=config.get_apikey(target))

    if username:
        _execute_push_login(
            client, image, username, password, email, reauth)
    image_name = utils.format_image_name(image, version)
    click.echo(f"Pushing {image_name} to the registry.")
    events = client.push(image_name, stream=True, decode=True)
    if utils.is_verbose():
        push_progress_cls = _LoggedPushProgress
    else:
        push_progress_cls = _PushProgress
    push_progress = push_progress_cls(events)
    push_progress.show()
    click.echo(f"The image {image_name} pushed successfully.")


def _execute_push_login(client, image, username, password, email, reauth):
    """Login if there're provided credentials for the registry"""
    registry = get_image_registry(image)
    resp = client.login(username=username, password=password,
                        email=email, registry=registry, reauth=reauth)
    if not (isinstance(resp, dict) and 'username' in resp or
            ('Status' in resp and resp['Status'] == 'Login Succeeded')):
        raise shub_exceptions.RemoteErrorException(
            "Docker registry login error.")
    click.echo(f"Login to {registry} succeeded.")


class _LoggedPushProgress(utils.BaseProgress):
    """Visualize push progress in verbose mode.

    Output all the events received from the docker daemon.
    """
    def handle_event(self, event):
        if 'error' in event and LOGIN_ERROR_MSG in event['error']:
            click.echo(
               "Something went wrong when trying to authenticate to Docker "
               "registry when pushing the image. Please ensure your "
               "credentials are correct and try again with --reauth flag.")
            raise shub_exceptions.RemoteErrorException(
                "Docker registry authentication error")
        super().handle_event(event)
        if 'status' in event:
            self.handle_status_event(event)

    def handle_status_event(self, event):
        msg = "Logs:{} {}".format(event['status'], event.get('progress'))
        utils.debug_log(msg)


class _PushProgress(_LoggedPushProgress):
    """Visualize push progress in non-verbose mode.

    Show total progress bar and separate bar for each pushed layer.
    """

    def __init__(self, push_events):
        super().__init__(push_events)
        # Total bar repesents total progress in terms of amount of layers.
        self.total_bar = self._create_total_bar()
        self.layers = set()
        # XXX: has to be OrderedDict to make tqdm.write/click.echo work as expected.
        # Otherwise it writes at random position, usually in the middle of the progress bars.
        self.layers_bars = OrderedDict()

    def handle_status_event(self, event):
        layer_id = event.get('id')
        status = event.get('status')
        progress = event.get('progressDetail')
        # `preparing` events are correlated with amount of layers to push
        if status in ('Preparing', 'Waiting'):
            self._add_layer(layer_id)
        # the events are final and used to update total bar once per layer
        elif status in ('Layer already exists', 'Pushed'):
            self._add_layer(layer_id)
            self.total_bar.update()
        # `pushing` events represents actual push process per layer
        elif event.get('status') == 'Pushing' and progress:
            progress_current = progress.get('current', 0)
            progress_total = max(progress.get('total', 0), progress_current)
            if layer_id not in self.layers_bars:
                if not progress_total:
                    return
                # create a progress bar per pushed layer
                self.layers_bars[layer_id] = self._create_bar_per_layer(
                    layer_id, progress_total, progress_current)
            bar = self.layers_bars[layer_id]
            bar.total = max(bar.total, progress_total)
            bar.update(max(progress_current - bar.n, 0))

    def _add_layer(self, layer_id):
        self.layers.add(layer_id)
        self.total_bar.total = max(self.total_bar.total, len(self.layers))
        self.total_bar.refresh()

    def show(self):
        super().show()
        self.total_bar.close()
        for bar in self.layers_bars.values():
            bar.close()

    def _create_total_bar(self):
        return utils.create_progress_bar(
            total=1,
            desc='Layers',
            # don't need rate here, let's simplify the bar
            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}'
        )

    def _create_bar_per_layer(self, layer_id, total, initial):
        return utils.create_progress_bar(
            desc=layer_id,
            total=total,
            initial=initial,
            unit='B',
            unit_scale=True,
            # don't need estimates here, keep only rate
            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt} [{rate_fmt}]',
        )



import unittest
from collections import namedtuple
from unittest import mock

from click.testing import CliRunner

from shub import fetch_eggs
from shub.exceptions import InvalidAuthException, RemoteErrorException

from .utils import AssertInvokeRaisesMixin, mock_conf


FakeResponse = namedtuple('FakeResponse', ['status_code'])


@mock.patch('shub.fetch_eggs.requests', autospec=True)
class FetchEggsTest(AssertInvokeRaisesMixin, unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.conf = mock_conf(self)

    def test_raises_auth_exception(self, requests_mock):
        fake_response = FakeResponse(403)
        requests_mock.get.return_value = fake_response
        self.assertInvokeRaises(InvalidAuthException, fetch_eggs.cli)

    def test_raises_exception_if_request_error(self, requests_mock):
        fake_response = FakeResponse(400)
        requests_mock.get.return_value = fake_response
        self.assertInvokeRaises(RemoteErrorException, fetch_eggs.cli)



import unittest
import textwrap
from unittest.mock import patch, MagicMock

import yaml
from click.testing import CliRunner
from yaml import CLoader as Loader

from shub import login
from shub.exceptions import AlreadyLoggedInException

from .utils import AssertInvokeRaisesMixin


VALID_KEY = 32 * '1'


@patch('shub.login.GLOBAL_SCRAPINGHUB_YML_PATH', new='.scrapinghub.yml')
@patch('shub.config.GLOBAL_SCRAPINGHUB_YML_PATH', new='.scrapinghub.yml')
@patch('shub.config.NETRC_PATH', new='.netrc')
@patch('shub.config.get_sources', new=MagicMock(return_value=[]))
class LoginTest(AssertInvokeRaisesMixin, unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def _run(self, user_input=VALID_KEY, files=None, fs=None, **kwargs):
        """Invokes the login cli on an isolated filesystem"""

        def write_local_test_files():
            for path, content in (files or {}).items():
                with open(path, 'w') as f:
                    f.write(content)

        def invoke():
            return self.runner.invoke(login.cli, input=user_input, **kwargs)

        def run():
            write_local_test_files()
            with patch.object(login, '_is_valid_apikey', return_value=True):
                return invoke()

        if fs:
            return run()

        with self.runner.isolated_filesystem() as fs:
            return run()

    def test_write_key_to_new_file(self):
        with self.runner.isolated_filesystem() as fs:
            self._run(fs=fs)
            with open('.scrapinghub.yml') as f:
                conf = yaml.load(f, Loader=Loader)
            self.assertEqual(conf['apikeys']['default'], VALID_KEY)

    def test_write_key_to_existing_file(self):
        VALID_SCRAPINGHUB_YML = textwrap.dedent("""
            endpoints:
                other: some_endpoint
        """)
        with self.runner.isolated_filesystem() as fs:
            files = {'.scrapinghub.yml': VALID_SCRAPINGHUB_YML}
            self._run(files=files, fs=fs)
            with open('.scrapinghub.yml') as f:
                conf = yaml.load(f, Loader=Loader)
            self.assertEqual(conf['apikeys']['default'], VALID_KEY)
            self.assertEqual(conf['endpoints']['other'], "some_endpoint")

    def test_suggest_project_key(self):
        PROJECT_SH_YML = textwrap.dedent("""
            apikeys:
                default: KEY_SUGGESTION
        """)
        files = {'scrapinghub.yml': PROJECT_SH_YML}
        result = self._run(files=files)
        err = 'Unexpected output: %s' % result.output
        self.assertTrue('KEY_SUGGESTION' in result.output, err)

    def test_suggest_env_key(self):
        result = self._run(env={'SHUB_APIKEY': 'SHUB_APIKEY_VALUE'})
        err = 'Unexpected output: %s' % result.output
        self.assertTrue('SHUB_APIKEY_VALUE' in result.output, err)

    def test_use_suggestion_to_log_in(self):
        apikey_suggestion = 'SHUB_APIKEY_VALUE'
        with self.runner.isolated_filesystem() as fs:
            self._run(
                env={'SHUB_APIKEY': apikey_suggestion},
                user_input='\n',
                fs=fs,
            )
            with open('.scrapinghub.yml') as f:
                conf = yaml.load(f, Loader=Loader)
            self.assertEqual(conf['apikeys']['default'], apikey_suggestion)

    def test_login_attempt_after_login_doesnt_lead_to_an_error(self):
        with self.runner.isolated_filesystem() as fs:
            self._run(fs=fs)
            self.assertInvokeRaises(AlreadyLoggedInException, login.cli,
                                    input=VALID_KEY)



__version__ = '2.16.0'


# Links to documentation to use over the project sources
DOCS_LINK = "https://shub.readthedocs.io/en/stable/"
DEPLOY_DOCS_LINK = DOCS_LINK + "deploying.html#deploying-dependencies"
CONFIG_DOCS_LINK = DOCS_LINK + "configuration.html"



import setuptools  # noqa: F401
import contextlib
import datetime
import errno
import json
import os
import subprocess
import sys
import re
import time

from collections import deque
from configparser import ConfigParser
from shutil import which
from packaging.version import Version
from glob import glob
from importlib import import_module
from tempfile import NamedTemporaryFile, TemporaryFile
from urllib.parse import urljoin

import click
import pip
import requests
import yaml
from click import ParamType

# https://github.com/scrapinghub/shub/pull/309#pullrequestreview-113977920
try:
    from pip import main as pip_main
except:  # noqa
    try:
        # For pip v20: https://tinyurl.com/pip20-error
        from pip._internal.cli.main import pip_main
    except ImportError:
        try:
            # For pip v9 and v10: https://tinyurl.com/y8mvl8rb
            from pip._internal.main import main as pip_main
        except ImportError:
            from pip._internal import main as pip_main

from scrapinghub import ScrapinghubClient, ScrapinghubAPIError, HubstorageClient

import shub
from shub.compat import to_native_str
from shub.exceptions import (
    BadParameterException, InvalidAuthException, NotFoundException,
    RemoteErrorException, SubcommandException, DeployRequestTooLargeException,
    print_warning,
)

SCRAPY_CFG_FILE = os.path.expanduser("~/.scrapy.cfg")
FALLBACK_ENCODING = 'utf-8'
STDOUT_ENCODING = sys.stdout.encoding or FALLBACK_ENCODING
LAST_N_LOGS = 30

# 50MB for a whole request, reserve 5KB for meta info (e.g. headers)
REQUEST_FILES_SIZE_LIMIT = 50 * 1024 * 1024 - 5 * 1024

_SETUP_PY_TEMPLATE = """\
# Automatically created by: shub deploy

from setuptools import setup, find_packages

setup(
    name         = 'project',
    version      = '1.0',
    packages     = find_packages(),
    entry_points = {'scrapy': ['settings = %(settings)s']},
)
"""


@contextlib.contextmanager
def remember_cwd():
    current_dir = os.getcwd()
    try:
        yield
    finally:
        os.chdir(current_dir)


def get_scrapinghub_client_from_config(conf):
    return ScrapinghubClient(
        conf.apikey, dash_endpoint=conf.endpoint
    )


def create_default_setup_py(**kwargs):
    closest = closest_file('scrapy.cfg')
    with remember_cwd():
        os.chdir(os.path.dirname(closest))
        if not os.path.exists('setup.py'):
            if 'settings' not in kwargs:
                kwargs['settings'] = get_config().get('settings', 'default')
            with open('setup.py', 'w') as f:
                f.write(_SETUP_PY_TEMPLATE % kwargs)
            click.echo(f"Created setup.py at {os.getcwd()}")


def make_deploy_request(url, data, files, auth, verbose, keep_log):
    _check_deploy_files_size(files)
    last_logs = deque(maxlen=LAST_N_LOGS)
    try:
        rsp = requests.post(url=url, auth=auth, data=data, files=files,
                            stream=True, timeout=300)
        rsp.raise_for_status()
        write_and_echo_logs(keep_log, last_logs, rsp, verbose)
        return True
    except requests.HTTPError as exc:
        rsp = exc.response

        if rsp.status_code == 403:
            raise InvalidAuthException

        try:
            error = rsp.json()['message']
            if 'Traceback' in error:
                error = ('\n---------- REMOTE TRACEBACK ----------\n' + error +
                         '\n---------- END OF REMOTE TRACEBACK ----------')
        except (ValueError, TypeError, KeyError):
            error = rsp.text or "Status %d" % rsp.status_code
        msg = f"Deploy failed ({rsp.status_code}):\n{error}"
        raise RemoteErrorException(msg)
    except requests.RequestException as exc:
        raise RemoteErrorException(f"Deploy failed: {exc}")


def _check_deploy_files_size(files):
    """Ensure that request's files total size is less than current limit."""
    ctx = click.get_current_context(silent=True)
    if not isinstance(files, list) or ctx and ctx.params.get('ignore_size'):
        return
    files_size = sum(
        len(fp) if isinstance(fp, str)
        else os.fstat(fp.fileno()).st_size
        for (fname, fp) in files
    )
    if files_size > REQUEST_FILES_SIZE_LIMIT:
        raise DeployRequestTooLargeException


def write_and_echo_logs(keep_log, last_logs, rsp, verbose):
    """It will write logs to temporal file and echo if verbose is True."""
    log_contents = b""
    for line in rsp.iter_lines():
        if verbose:
            click.echo(line)
        last_logs.append(line)
        log_contents += line + b'\n'
    deployed = _is_deploy_successful(last_logs)
    if not deployed:
        keep_log = True
    echo_short_log_if_deployed(deployed, last_logs, verbose=verbose)

    with NamedTemporaryFile(prefix='shub_deploy_', suffix='.log',
                            delete=not keep_log) as log_file:
        log_file.write(log_contents)
        if keep_log:
            click.echo("Deploy log location: %s" % log_file.name)
        if not deployed:
            try:
                last_log = last_logs[-1]
            except IndexError:
                last_log = "(No log messages)"
            raise RemoteErrorException(f"Deploy failed: {last_log}")


def echo_short_log_if_deployed(deployed, last_logs, log_file=None, verbose=False):
    if deployed:
        if not verbose:
            click.echo(last_logs[-1])
    else:
        if not verbose:
            click.echo("Deploy log last %s lines:" % len(last_logs))
            for line in last_logs:
                click.echo(line)


def _is_deploy_successful(last_logs):
    try:
        data = json.loads(to_native_str(last_logs[-1]))
        if 'status' in data and data['status'] == 'ok':
            return True
    except Exception:
        pass


@contextlib.contextmanager
def patch_sys_executable():
    """
    Context manager that monkey-patches sys.executable to point to the Python
    interpreter.

    Some scripts, in particular pip, depend on sys.executable pointing to the
    Python interpreter. When frozen, however, sys.executable points to the
    stand-alone file (i.e. the frozen script).
    """
    if getattr(sys, 'frozen', False):
        orig_exe = sys.executable
        py_exe = find_exe('python')
        # PyInstaller sets this environment variable in its bootloader. Remove
        # it so the system-wide Python installation uses its own library path
        # (this is particularly important if the system Python version differs
        # from the Python version that the binary was compiled with)
        orig_lib_path = os.environ.pop('LD_LIBRARY_PATH', None)
        sys.executable = py_exe
        yield
        sys.executable = orig_exe
        if orig_lib_path:
            os.environ['LD_LIBRARY_PATH'] = orig_lib_path
    else:
        yield


def find_exe(exe_name):
    exe = which(exe_name)
    if not exe:
        raise NotFoundException(f"Please install {exe_name}")
    return exe


def run_cmd(*args, **kwargs):
    """Run a command and return its output, decoded by the stdout encoding and
    stripped of trailing newlines. `args` and `kwargs` are forwarded to
    `subprocess.check_output`.

    Raises SubcommandException on non-zero exit codes or other subprocess
    errors."""
    def _clean(s):
        return s.decode(STDOUT_ENCODING).replace(os.linesep, '\n').strip('\n')

    with TemporaryFile() as tmpfile:
        kwargs.setdefault('stderr', tmpfile)
        try:
            return _clean(subprocess.check_output(*args, **kwargs))
        except subprocess.CalledProcessError as e:
            msg = ("Error while calling subcommand: %s\n\nCOMMAND OUTPUT\n"
                   "--------------\n%s" % (e, _clean(e.output)))
            tmpfile.seek(0)
            e.stderr = _clean(tmpfile.read())
            if e.stderr:
                msg += "\n\nSTDERR\n------\n%s" % e.stderr
            raise SubcommandException(msg)


def pwd_version():
    """
    Try to find version information on whatever lives in the current directory
    -- most commonly a Python package or Scrapy project -- by trying (in this
    order):
        - git commit/branch
        - mercurial commit/branch
        - bazaar commit/branch
        - setup.py in this folder
        - setup.py next to closest scrapy.cfg
    If none of these work, fall back to the UNIX time.
    """
    ver = pwd_git_version()
    if not ver:
        ver = pwd_hg_version()
    if not ver:
        ver = pwd_bzr_version()
    if not ver and os.path.isfile('setup.py'):
        ver = _last_line_of(run_python(['setup.py', '--version']))
    if not ver:
        closest_scrapycfg = closest_file('scrapy.cfg')
        if closest_scrapycfg:
            setuppy = os.path.join(os.path.dirname(closest_scrapycfg),
                                   'setup.py')
            if os.path.isfile(setuppy):
                ver = _last_line_of(run_python([setuppy, '--version']))
    if not ver:
        ver = str(int(time.time()))
    ver = re.sub(r'[^\w.-]+', '', ver)
    return ver


def pwd_git_version():
    git = which('git')
    if not git:
        return None
    try:
        commit_id = run_cmd([git, 'describe', '--always'])
    except SubcommandException:
        try:
            commit_id = run_cmd([git, 'rev-list', '--count', 'HEAD'])
        except SubcommandException:
            return None
    branch = run_cmd([git, 'rev-parse', '--abbrev-ref', 'HEAD'])
    return f'{commit_id}-{branch}'


def pwd_hg_version():
    hg = which('hg')
    if not hg:
        return None
    try:
        commit_id = run_cmd([hg, 'tip', '--template', '{rev}'])
    except SubcommandException:
        return None
    branch = run_cmd([hg, 'branch'])
    return f'r{commit_id}-{branch}'


def pwd_bzr_version():
    bzr = which('bzr')
    if not bzr:
        return None
    try:
        return '%s' % run_cmd([bzr, 'revno']).strip()
    except SubcommandException:
        return None


def run_python(cmd, *args, **kwargs):
    """
    Call Python interpreter with supplied list of arguments and return its
    output. `args` and `kwargs` are forwarded to `subprocess.check_output`.
    """
    with patch_sys_executable():
        return run_cmd([sys.executable] + cmd, *args, **kwargs)


def decompress_egg_files(directory=None):
    try:
        EXTS = pip.utils.ARCHIVE_EXTENSIONS
    except AttributeError:
        try:
            EXTS = pip._internal.utils.misc.ARCHIVE_EXTENSIONS
        except AttributeError:
            EXTS = ('.zip', '.whl', '.tar', '.tar.gz', '.tar.bz2')
    try:
        unpack_file = pip.utils.unpack_file
    except AttributeError:
        # XXX a work-around for pip >= 10.0
        try:
            unpack_file = pip.util.unpack_file
        except AttributeError:
            try:
                unpack_file = pip._internal.utils.misc.unpack_file
            except AttributeError:
                from pip._internal.utils.unpacking import unpack_file
    pathname = "*"
    if directory is not None:
        pathname = os.path.join(directory, pathname)
    eggs = [f for ext in EXTS for f in glob(pathname + "%s" % ext)]
    if not eggs:
        files = glob(pathname)
        err = ('No egg files with a supported file extension were found. '
               'Files: %s' % ', '.join(files))
        raise NotFoundException(err)
    for egg in eggs:
        click.echo("Uncompressing: %s" % egg)
        egg_ext = EXTS[list(egg.endswith(ext) for ext in EXTS).index(True)]
        decompress_location = egg[:-len(egg_ext)]
        try:
            unpack_file(egg, decompress_location, None)
        except TypeError:
            unpack_file(egg, decompress_location, None, None)


def build_and_deploy_eggs(project, endpoint, apikey):
    egg_dirs = (f for f in glob('*') if os.path.isdir(f))

    for egg_dir in egg_dirs:
        os.chdir(egg_dir)
        build_and_deploy_egg(project, endpoint, apikey)
        os.chdir('..')


def build_and_deploy_egg(project, endpoint, apikey):
    """Builds and deploys the current dir's egg"""
    click.echo("Building egg in: %s" % os.getcwd())
    try:
        run_python(['setup.py', 'bdist_egg'])
    except SubcommandException:
        # maybe a C extension or distutils package, forcing bdist_egg
        click.echo("Couldn't build an egg with vanilla setup.py, trying with "
                   "setuptools...")
        script = "import setuptools; __file__='setup.py'; execfile('setup.py')"
        run_python(['-c', script, 'bdist_egg'])

    _deploy_dependency_egg(project, endpoint, apikey)


def _deploy_dependency_egg(project, endpoint, apikey, name=None, version=None, egg_info=None):
    name = name or _get_dependency_name()
    version = version or pwd_version()
    egg_info = egg_info or _get_egg_info(name)
    egg_name, egg_path = egg_info
    url = urljoin(endpoint, 'eggs/add.json')
    data = {'project': project, 'name': name, 'version': version}
    auth = (apikey, '')

    click.echo(f'Deploying dependency {name} {version} to Scrapy Cloud project {project}')

    with open(egg_path, 'rb') as egg_fp:
        files = {'egg': (egg_name, egg_fp)}
        make_deploy_request(url, data, files, auth, False, False)

    success = "Deployed eggs list at: https://app.zyte.com/p/%s/deploy/"
    click.echo(success % project)


def _last_line_of(s):
    return s.split('\n')[-1]


def _get_dependency_name():
    # In some cases, python setup.py --name returns more than one line, so we
    # use the last one to get the name
    return _last_line_of(run_python(['setup.py', '--name']))


def _get_egg_info(name):
    egg_filename = name.replace('-', '_')
    egg_path_glob = os.path.join('dist', '%s*' % egg_filename)
    egg_path = glob(egg_path_glob)[0]
    return egg_filename, egg_path


def get_job_specs(job):
    """
    Parse job identifier into valid job id and corresponding API key.

    With projects default=10 and external=20 defined in config:
    * 1/1 -> 10/1/1
    * 2/2/2 -> 2/2/2
    * external/2/2 -> 20/2/2

    It also accepts job URLs from Scrapinghub.
    """
    match = re.match(r'^((\w+)/)?(\d+/\d+)$', job)
    if not match:
        job_url_re = r'^https?://[^/]+/p/((\d+)/)(?:job/)?(\d+/\d+).*'
        match = re.match(job_url_re, job)
    if not match:
        raise BadParameterException(
            "Job ID {} is invalid. Format should be spiderid/jobid (inside a "
            "project) or target/spiderid/jobid, where target can be either a "
            "project ID or an identifier defined in scrapinghub.yml."
            "".format(job),
            param_hint='job_id',
        )
    # XXX: Lazy import due to circular dependency
    from shub.config import get_target_conf
    targetconf = get_target_conf(match.group(2) or 'default')
    return (f"{targetconf.project_id}/{match.group(3)}",
            targetconf.apikey)


def get_job(job):
    jobid, apikey = get_job_specs(job)
    hsc = HubstorageClient(auth=apikey)
    job = hsc.get_job(jobid)
    if not job.metadata:
        raise NotFoundException(f'Job {jobid} does not exist')
    return job


def closest_file(filename, path='.', prevpath=None):
    """
    Return the path to the closest file with the given filename by traversing
    the current directory and its parents
    """
    if path == prevpath:
        return None
    path = os.path.abspath(path)
    thisfile = os.path.join(path, filename)
    if os.path.exists(thisfile):
        return thisfile
    return closest_file(filename, os.path.dirname(path), path)


def inside_project():
    scrapy_module = os.environ.get('SCRAPY_SETTINGS_MODULE')
    if scrapy_module is not None:
        try:
            import_module(scrapy_module)
        except ImportError as exc:
            print_warning("Cannot import scrapy settings module %s: %s"
                          "" % (scrapy_module, exc))
        else:
            return True
    return bool(closest_file('scrapy.cfg'))


def get_config(use_closest=True):
    """Get Scrapy config file as a ConfigParser"""
    sources = get_sources(use_closest)
    cfg = ConfigParser()
    cfg.read(sources)
    return cfg


def get_sources(use_closest=True):
    xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or \
        os.path.expanduser('~/.config')
    sources = ['/etc/scrapy.cfg', r'c:\scrapy\scrapy.cfg',
               xdg_config_home + '/scrapy.cfg',
               os.path.expanduser('~/.scrapy.cfg')]
    if use_closest:
        closest_scrapy_cfg_path = closest_file('scrapy.cfg')
        if closest_scrapy_cfg_path:
            sources.append(closest_scrapy_cfg_path)
    return sources


def get_scrapycfg_targets(cfgfiles=None):
    cfg = ConfigParser()
    cfg.read(cfgfiles or [])
    baset = dict(cfg.items('deploy')) if cfg.has_section('deploy') else {}
    targets = {}
    targets['default'] = baset
    for x in cfg.sections():
        if x.startswith('deploy:'):
            t = baset.copy()
            t.update(cfg.items(x))
            targets[x[7:]] = t
    for tname, t in list(targets.items()):
        try:
            int(t.get('project', 0))
        except ValueError:
            # Don't import non-numeric project IDs, and also throw away the
            # URL and credentials associated with these projects (since the
            # project ID does not belong to SH, neither do the endpoint or the
            # auth information)
            del targets[tname]
        if t.get('url', "").endswith('scrapyd/'):
            t['url'] = t['url'][:-8]
    targets.setdefault('default', {})
    return targets


def job_live(job, refresh_meta_after=60):
    """
    Check whether job is in 'pending' or 'running' state. If job metadata was
    fetched longer than `refresh_meta_after` seconds ago, refresh it.
    """
    if not hasattr(job, '_metadata_updated'):
        # Assume just loaded
        job._metadata_updated = time.time()
    if time.time() - job._metadata_updated > refresh_meta_after:
        job.metadata.expire()
        # Fetching actually happens on job.metadata['state'], but close enough
        job._metadata_updated = time.time()
    return job.metadata['state'] in ('pending', 'running')


def job_resource_iter(job, resource, output_json=False, follow=True,
                      tail=None):
    """
    Given a python-hubstorage job and resource (e.g. job.items), return a
    generator that periodically checks the job resource and yields its items.
    The generator will exit when the job has finished.

    As a handy shortcut, iter_func will be iterated through only once if
    `follow` is set to `False`.
    """
    last_item_key = None
    if tail is not None:
        total_nr_items = resource.stats()['totals']['input_values']
        # This is the last entry to be skipped, i.e. it will NOT be displayed
        last_item = total_nr_items - tail - 1
        if last_item >= 0:
            last_item_key = f'{job.key}/{last_item}'
    if not job_live(job):
        follow = False
    resource_iter = resource.iter_json if output_json else resource.iter_values
    if not follow:
        for item in resource_iter(startafter=last_item_key):
            yield item
        return
    while True:
        # XXX: Always use iter_json until Kumo team fixes iter_values to also
        # return '_key'
        for json_line in resource.iter_json(startafter=last_item_key):
            item = json.loads(json_line)
            last_item_key = item['_key']
            yield json_line if output_json else item
        if not job_live(job):
            break
        # Workers only upload data to hubstorage every 15 seconds
        time.sleep(15)


def latest_github_release(force_update=False, timeout=1., cache=None):
    """
    Get GitHub data for latest shub release. If it was already requested today,
    return a cached version unless ``force_update`` is set to ``True``.
    """
    REQ_URL = "https://api.github.com/repos/scrapinghub/shub/releases/latest"
    cache = cache or os.path.join(click.get_app_dir('scrapinghub'),
                                  'last_release.txt')
    today = datetime.date.today().toordinal()
    if not force_update and os.path.isfile(cache):
        with open(cache) as f:
            try:
                release_data = json.load(f)
            except Exception:
                release_data = {}
        # Check for equality (and not smaller or equal) so we don't get thrown
        # off track if the clock was ever misconfigured and a future date was
        # saved
        if release_data.get('_shub_last_update', 0) == today:
            return release_data
    release_data = requests.get(REQ_URL, timeout=timeout).json()
    release_data['_shub_last_update'] = today
    try:
        shubdir = os.path.dirname(cache)
        try:
            os.makedirs(shubdir)
        except OSError:
            if not os.path.isdir(shubdir):
                raise
        with open(cache, 'w') as f:
            json.dump(release_data, f)
    except Exception:
        pass
    return release_data


def update_available(silent_fail=True):
    """
    Check whether most recent GitHub release of shub is newer than the shub
    version in use. If a newer version is available, return a link to the
    release on GitHub, otherwise return ``None``.
    """
    try:
        release_data = latest_github_release()
        latest_rls = Version(release_data['name'].lstrip('v'))
        used_rls = Version(shub.__version__)
        if used_rls >= latest_rls:
            return None
        return release_data['html_url']
    except Exception:
        if not silent_fail:
            raise
        # Don't let this interfere with shub usage
        return None


def download_from_pypi(dest, pkg=None, reqfile=None, extra_args=None):
    if (not pkg and not reqfile) or (pkg and reqfile):
        raise ValueError('Call with either pkg or reqfile')
    extra_args = extra_args or []
    pip_version = Version(getattr(pip, '__version__', '1.0'))
    cmd = 'install'
    no_wheel = []
    target = [pkg] if pkg else ['-r', reqfile]
    if pip_version >= Version('1.4'):
        no_wheel = ['--no-use-wheel']
    if pip_version >= Version('7'):
        no_wheel = ['--no-binary=:all:']
    if pip_version >= Version('8'):
        cmd = 'download'
    with patch_sys_executable():
        pip_main([cmd, '-d', dest, '--no-deps'] + no_wheel + extra_args +
                 target)


@contextlib.contextmanager
def update_yaml_dict(conf_path=None):
    """
    Context manager for updating a YAML file. Key ordering and comments are not
    preserved.
    """
    if not conf_path:
        click.secho("Using update_yaml_dict without path is deprecated. Import"
                    " GLOBAL_SCRAPINGHUB_YML_PATH from shub.config",
                    fg='yellow')
        from shub.config import GLOBAL_SCRAPINGHUB_YML_PATH
        conf_path = GLOBAL_SCRAPINGHUB_YML_PATH
    try:
        with open(conf_path) as f:
            conf = yaml.safe_load(f) or {}
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        conf = {}
    # Code inside context manager is executed after this yield
    yield conf
    # Avoid writing "key: {}"
    for key in list(conf):
        if conf[key] == {}:
            del conf[key]
    with open(conf_path, 'w') as f:
        # Avoid writing "{}"
        if conf:
            yaml.safe_dump(conf, f, default_flow_style=False)


def has_project_access(project, endpoint, apikey):
    """Check whether an API key has access to a given project. May raise
    InvalidAuthException if the API key is invalid (but not if it is valid but
    lacks access to the project)"""
    client = ScrapinghubClient(apikey, dash_endpoint=endpoint)
    try:
        return project in client.projects.list()
    except ScrapinghubAPIError as e:
        if 'Authentication failed' in str(e):
            raise InvalidAuthException
        else:
            raise RemoteErrorException(str(e))


def get_project_dir():
    """Get the path to the closest directory that contains either
    ``scrapinghub.yml``. ``scrapy.cfg``, or ``Dockerfile`` (in this priority).
    """
    for filename in ['scrapinghub.yml', 'scrapy.cfg', 'Dockerfile']:
        closest = closest_file(filename)
        if closest:
            return os.path.dirname(closest)
    raise NotFoundException(
        "Cannot find project: There is no scrapinghub.yml, scrapy.cfg, or "
        "Dockerfile in this directory or any of the parent directories.")


def _get_target_project(conf, target):
    """Ask for project ID (or use given target if numerical) and confirm that
    user is logged in and has project access"""
    # Will raise MissingAuthException if user is not logged in
    endpoint, apikey = conf.get_endpoint(0), conf.get_apikey(0)
    if target.isdigit():
        project = int(target)
        target = 'default'
    else:
        project = click.prompt("Target project ID", type=int)
    if not has_project_access(project, endpoint, apikey):
        raise InvalidAuthException(
            "The account you logged in to has no access to project {}. "
            "Please double-check the project ID and make sure you logged "
            "in to the correct acount.".format(project),
        )
    return target, project


def _detect_custom_image_project():
    """Guess if the user may want to deploy a custom image based on the
    existence of ``scrapy.cfg`` and ``Dockerfile``. If there are both, ask."""
    project_dir = get_project_dir()
    has_scrapy_cfg = os.path.exists(os.path.join(project_dir, 'scrapy.cfg'))
    has_dockerfile = os.path.exists(os.path.join(project_dir, 'Dockerfile'))
    if has_scrapy_cfg and has_dockerfile:
        return click.confirm(
            "You have a Dockerfile in your project directory. Would you like "
            "to deploy it as custom image?", default=True)
    elif has_dockerfile:
        return True
    return False


def _update_conf(conf, target, project, repository):
    """Update configuration target with given ``project`` and ``repository``"""
    if project:
        # XXX: Save {'id': project} once we normalize project config on loading
        conf.projects[target] = project
    if repository:
        if target == 'default':
            conf.images[target] = repository
        else:
            # XXX: Remove once we normalize project config on loading
            if not isinstance(conf.projects[target], dict):
                conf.projects[target] = {'id': conf.projects[target]}
            conf.projects[target]['image'] = repository


def _update_conf_file(filename, target, project, repository):
    """Load the given config file, update ``target`` with the given ``project``
    and ``repository``, then save it. If the file does not exist, it will be
    created."""
    try:
        # XXX: Runtime import to avoid circular dependency
        from shub.config import ShubConfig
        conf = ShubConfig()
        if os.path.exists(filename):
            conf.load_file(filename)
        _update_conf(conf, target, project, repository)
        conf.save(filename)
    except Exception as e:
        click.echo(
            "There was an error while trying to write to %s: %s"
            "" % (filename, e),
        )
    else:
        click.echo("Saved to %s." % filename)


class _AnyParamType(ParamType):
    name = "any"

    def convert(self, value, param, ctx):
        return value


def create_scrapinghub_yml_wizard(conf, target='default', image=None):
    """
    Ask user for project ID, ensure they have access to that project, and save
    it in the local ``scrapinghub.yml``.

    If ``image`` is ``True``, the user will also be asked for the image
    repository to use. If ``image`` is ``None``, the wizard will ask for a
    repository if ``Dockerfile`` exists.

    The wizard will only ever ask questions and touch the configuration if at
    least one of these two conditions is met:

        1. There is no local scrapinghub.yml
        2. ``image`` is ``True``, and the given ``target`` is defined but has
           no image repository configured; this will happen when transitioning
           a previously stack-based project to a custom image-based one, i.e.
           when you already have a ``scrapinghub.yml`` but now run ``shub image
           build`` for the first time

    In all other cases, the wizard will return without asking questions and
    without altering ``conf``.
    """
    closest_sh_yml = os.path.join(get_project_dir(), 'scrapinghub.yml')
    run_wizard = (
        not os.path.exists(closest_sh_yml) or
        (image and target in conf.projects
            and not conf.get_target_conf(target).image)
    )
    if not run_wizard:
        return
    project = None
    repository = None
    if target not in conf.projects and 'default' not in conf.projects:
        target, project = _get_target_project(conf, target)
        if target == 'default':
            click.echo(
                "Saving project %d as default target. You can deploy to it "
                "via 'shub deploy' from now on" % project)
        else:
            click.echo(
                "Saving project %d as target '%s'. You can deploy to it via "
                "'shub deploy %s' from now on" % (project, target, target))
    if image or (image is None and _detect_custom_image_project()):
        repository = click.prompt(
            "Image repository (leave empty to use Scrapinghub's repository)",
            default=True, show_default=False, type=_AnyParamType())
    _update_conf(conf, target, project, repository)
    _update_conf_file(closest_sh_yml, target, project, repository)



import json

import click
import docker
import requests
from urllib.parse import urljoin

from shub.exceptions import ShubException
from shub.config import load_shub_config, list_targets_callback
from shub.image import utils


SETTING_TYPES = ['project_settings',
                 'organization_settings',
                 'enabled_addons']

SHORT_HELP = 'List spiders.'

HELP = """
List command tries to run your image locally and get a spiders list.

Internally, this command is a simple wrapper to `docker run` and uses
docker daemon on your system to run a new container using your image.
Before creating the container, there's a Dash call to get your project
settings to get your spiders list properly (respecting SPIDERS_MODULE
setting, etc).

Image should be set via scrapinghub.yml, section "images". If version is not
provided, the tool uses VCS-based stamp over project directory (the same as
shub utils itself).
"""


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True, help="stream logs to console")
@click.option("-s", "--silent", is_flag=True,
              help="don't warn if Dash project is not defined in config")
@click.option("-V", "--version", help="release version")
def cli(target, debug, verbose, silent, version):
    list_cmd_full(target, silent, version)


def list_cmd_full(target, silent, version):
    config = load_shub_config()
    image = config.get_image(target)
    version = version or config.get_version()
    image_name = utils.format_image_name(image, version)
    target_conf = config.get_target_conf(target)
    metadata = list_cmd(image_name,
                        target_conf.project_id,
                        target_conf.endpoint,
                        target_conf.apikey)
    for spider in metadata.get('spiders', []):
        click.echo(spider)


def list_cmd(image_name, project, endpoint, apikey):
    """Short version of list cmd to use with deploy cmd."""
    settings = _get_project_settings(project, endpoint, apikey)
    environment = {'JOB_SETTINGS': json.dumps(settings)}
    exit_code, logs = _run_cmd_in_docker_container(
        image_name, 'shub-image-info', environment)
    if exit_code == 0:
        return _extract_metadata_from_image_info_output(logs)
    # shub-image-info command not found, fallback to list-spiders
    elif exit_code == 127:
        # FIXME we should pass some value for SCRAPY_PROJECT_ID anyway
        # to handle `scrapy list` cmd properly via sh_scrapy entrypoint
        # environment['SCRAPY_PROJECT_ID'] = str(project) if project else ''
        exit_code, logs = _run_cmd_in_docker_container(
            image_name, 'list-spiders', environment)
        if exit_code != 0:
            click.echo(logs)
            raise ShubException('Container with list cmd exited with code %s' % exit_code)
        return {
            'project_type': 'scrapy',
            'spiders': utils.valid_spiders(logs.splitlines()),
        }
    else:
        click.echo(logs)
        raise ShubException(
            'Container with shub-image-info cmd exited with code %s' % exit_code)


def _get_project_settings(project, endpoint, apikey):
    utils.debug_log(f'Getting settings for {project} project:')
    req = requests.get(
        urljoin(endpoint, '/api/settings/get.json'),
        params={'project': project},
        auth=(apikey, ''),
        timeout=300,
        allow_redirects=False
    )
    req.raise_for_status()
    utils.debug_log(f"Response: {req.json()}")
    return {k: v for k, v in req.json().items() if k in SETTING_TYPES}


def _run_cmd_in_docker_container(image_name, command, environment):
    """Run a command inside the image container."""
    client = utils.get_docker_client()
    container = client.create_container(
        image=image_name,
        command=[command],
        environment=environment,
    )
    if 'Id' not in container:
        raise ShubException("Create container error:\n %s" % container)
    try:
        client.start(container)
    except docker.errors.APIError as e:
        explanation = utils.ensure_unicode(e.explanation or '')
        if 'executable file not found' in explanation:
            # docker.errors.APIError: 500 Server Error:
            # Internal Server Error ("Cannot start container xxx:
            # [8] System error: exec: "shub-image-info":
            # executable file not found in $PATH")
            return 127, None
        raise
    statuscode = client.wait(container=container['Id'])['StatusCode']
    logs = client.logs(
        container=container['Id'], stream=False, timestamps=False,
        stdout=True, stderr=True if statuscode else False,
    )
    return statuscode, utils.ensure_unicode(logs)


def _extract_metadata_from_image_info_output(output):
    """Extract and validate spiders list from `shub-image-info` output."""

    def raise_shub_image_info_error(error):
        """Helper to raise ShubException with prefix and output"""
        msg = f"shub-image-info: {error} \n[output '{output}']"
        raise ShubException(msg)

    try:
        metadata = json.loads(output)
        project_type = metadata.get('project_type')
    except (AttributeError, ValueError):
        raise_shub_image_info_error('output is not a valid JSON dict')
    if not isinstance(project_type, str):
        raise_shub_image_info_error('"project_type" key is required and must be a string')

    spiders_list = metadata.get('spiders')
    if not isinstance(spiders_list, list):
        raise_shub_image_info_error('"spiders" key is required and must be a list')
    spiders, scripts = [], []
    for name in spiders_list:
        if not (name and isinstance(name, str)):
            raise_shub_image_info_error("spider name can't be empty or non-string")
        if project_type == 'scrapy' and name.startswith('py:'):
            scripts.append(name[3:])
        else:
            spiders.append(name)
    return {
        'project_type': project_type,
        'spiders': utils.valid_spiders(spiders),
        'scripts': utils.valid_spiders(scripts),
    }



import os
import textwrap
from string import Template

import click

from shub import exceptions as shub_exceptions
from shub import utils as shub_utils


DOCKER_APP_DIR = '/app'
DOCKERFILE_TEMPLATE = """\
FROM $base_image
$system_deps
$system_env
RUN mkdir -p {docker_app_dir}
WORKDIR {docker_app_dir}
$requirements
COPY . {docker_app_dir}
RUN python -m pip install .
""".format(docker_app_dir=DOCKER_APP_DIR)

DEFAULT_BASE_IMAGE = "scrapinghub/scrapinghub-stack-scrapy:2.13-20250721"
RECOMMENDED_PYTHON_DEPS = [
    'guppy==0.1.10',
]

SHORT_HELP = "Create Dockerfile for existing Scrapy project."

HELP = """
Init command creates a Dockerfile for existing Scrapy project. This tool is for users
who want to create a custom Docker image and don't have a Dockerfile yet. If generated
Dockerfile doesn't fit your project feel free to edit it.

Python packages

If there's a requirements.txt file in the project directory - it will be added to the
Dockerfile. Also it's possible to provide a path to requirements file via --requirements
option. Otherwise new requirements.txt file will be created in the project directory
with the recommended Python packages. Use --list-recommended-reqs to list them.

It's recommended to include scrapinghub-entrypoint-scrapy package - it is a
support layer that passes data from the job to Scrapinghub storage. Otherwise
you will need to send data to Scrapinghub storage using HTTP API.

System packages

You can extend list of system packages installed in the image via --add-deps option.
"""


def list_recommended_python_reqs(ctx, param, value):
    """List recommended Python requirements"""
    if not value:
        return
    click.echo("Recommended Python deps list:")
    for dep in RECOMMENDED_PYTHON_DEPS:
        click.echo(f'- {dep}')
    ctx.exit()


def _deprecate_base_deps_parameter(ctx, param, value):
    if value:
        click.echo("WARNING: --base-deps parameter is deprecated. "
                   "Please use --add-deps parameter instead.",
                   err=True)
    return value


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option("--list-recommended-reqs", is_flag=True, is_eager=True,
              expose_value=False, callback=list_recommended_python_reqs,
              help="list recommended python requirements")
@click.option("--project", default="default",
              help="project name to get settings module from scrapy.cfg")
@click.option("--base-image", default=DEFAULT_BASE_IMAGE,
              help="base docker image name")
@click.option("--base-deps", default='',
              help="[DEPRECATED] a comma-separated list with base system dependencies",
              callback=_deprecate_base_deps_parameter)
@click.option("--add-deps",
              help="a comma-separated list with additional system dependencies")
@click.option("--requirements", default="requirements.txt",
              help="path to requirements.txt")
def cli(project, base_image, base_deps, add_deps, requirements):
    closest_scrapy_cfg = shub_utils.closest_file('scrapy.cfg')
    scrapy_config = shub_utils.get_config()
    if not closest_scrapy_cfg or not scrapy_config.has_option('settings', project):
        raise shub_exceptions.BadConfigException(
            'Cannot find Scrapy project settings. Please ensure that current directory '
            'contains scrapy.cfg with settings section, see example at '
            'https://doc.scrapy.org/en/latest/topics/commands.html#default-structure-of-scrapy-projects')  # NOQA
    project_dir = os.path.dirname(closest_scrapy_cfg)
    dockefile_path = os.path.join(project_dir, 'Dockerfile')
    if os.path.exists(dockefile_path):
        raise shub_exceptions.ShubException('Found a Dockerfile in the project directory, aborting')
    settings_module = scrapy_config.get('settings', 'default')
    shub_utils.create_default_setup_py(settings=settings_module)
    values = {
        'base_image':   base_image,
        'system_deps':  _format_system_deps(base_deps, add_deps),
        'system_env':   _format_system_env(settings_module),
        'requirements': _format_requirements(project_dir, requirements),
    }
    values = {key: value if value else '' for key, value in values.items()}
    source = Template(DOCKERFILE_TEMPLATE)
    results = source.substitute(values)
    results = results.replace('\n\n', '\n')
    with open(dockefile_path, 'w') as dockerfile:
        dockerfile.write(results)
    click.echo(f"Dockerfile is saved to {dockefile_path}")


def _format_system_deps(base_deps, add_deps):
    """Prepare a list with system dependencies install cmds"""
    system_deps = base_deps.split(',') if base_deps != '-' else []
    if add_deps:
        system_add_deps = add_deps.split(',')
        system_deps = list(set(system_deps + system_add_deps))
    system_deps = sorted(filter(None, system_deps))
    if not system_deps:
        return
    commands = ["apt-get update -qq",
                "apt-get install -qy {}".format(' '.join(system_deps)),
                "rm -rf /var/lib/apt/lists/*"]
    return 'RUN ' + ' && \\\n    '.join(
        [_wrap(cmd) for cmd in commands])


def _wrap(text):
    """Wrap dependencies with separator"""
    lines = textwrap.wrap(text, subsequent_indent='    ',
                          break_long_words=False,
                          break_on_hyphens=False)
    return ' \\\n'.join(lines)


def _format_system_env(settings_module):
    rows = ['ENV TERM xterm']
    if settings_module:
        rows.append('ENV SCRAPY_SETTINGS_MODULE %s' % settings_module)
    return '\n'.join(rows)


def _format_requirements(project_dir, requirements):
    """Prepare cmds for project requirements"""
    rel_reqs_path = os.path.relpath(
        os.path.join(project_dir, requirements), project_dir)
    if os.path.isfile(rel_reqs_path):
        if rel_reqs_path.startswith('../'):
            raise shub_exceptions.BadParameterException(
                "Requirements file must be inside your project directory, "
                "otherwise it will not be included in the Docker build context.")
    else:
        # let's create requirements.txt with base dependencies
        with open(rel_reqs_path, 'w') as reqs_file:
            reqs_file.writelines("%s\n" % line for line in RECOMMENDED_PYTHON_DEPS)
        click.echo('Created base requirements.txt in project dir.')
    rows = [
        f'COPY ./{rel_reqs_path} {DOCKER_APP_DIR}/requirements.txt',
        'RUN pip install --no-cache-dir -r requirements.txt',
    ]
    return '\n'.join(rows)



import click

from shub import exceptions as shub_exceptions
from shub.config import load_shub_config, list_targets_callback
from shub.image import utils

SHORT_HELP = "Test a built image with Scrapy Cloud contract"
HELP = """
A command to test an image after build step to make sure it fits contract.

It consists of the following steps:

1) check that image exists on local machine
2) check that image has start-crawl entrypoint
3) check that image has shub-image-info entrypoint

If any of the checks fails - the test command fails as a whole. By default,
the test command is also executed automatically as a part of build command
in its end (if you do not provide -S/--skip-tests parameter explicitly).
"""

IMAGE_SIZE_LIMIT = 3 * 1024 * 1024 * 1024  # 3GB
CONTRACT_CMD_NOT_FOUND_WARNING = (
    'Command %s is not found in the image. '
    'Please make sure you provided it according to Scrapy Cloud contract '
    '(https://shub.readthedocs.io/en/stable/custom-images-contract.html) or '
    'added scrapinghub-entrypoint-scrapy>=0.8.0 to your requirements file '
    'if you use Scrapy.'
)
LIST_SPIDERS_DEPRECATED_WARNING = (
    'list-spiders command is deprecated in favour of shub-image-info command: '
    'its format is described well in Scrapy Cloud contract '
    '(https://shub.readthedocs.io/en/stable/custom-images-contract.html), '
    'please review and update your code.'
)
IMAGE_TOO_LARGE_WARNING = (
    'Custom image for the project is too large (more than 3GB), it can lead '
    'to various performance issues when running it in Scrapy Cloud. '
    'Please reduce the image size or ask support team for help '
    '(one of the recommended articles to start with is '
    'https://www.codacy.com/blog/five-ways-to-slim-your-docker-images/).'
)


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True,
              help="stream test logs to console")
@click.option("-V", "--version", help="release version")
def cli(target, debug, verbose, version):
    test_cmd(target, version)


def test_cmd(target, version):
    config = load_shub_config()
    image = config.get_image(target)
    version = version or config.get_version()
    image_name = utils.format_image_name(image, version)
    docker_client = utils.get_docker_client()
    for check in [_check_image_size,
                  _check_start_crawl_entry,
                  _check_shub_image_info_entry]:
        check(image_name, docker_client)


def _check_image_size(image_name, docker_client):
    """Check that the image exists on local machine and validate its size."""
    # if there's no docker lib, the command will fail earlier
    # with an exception when getting a client in get_docker_client()
    from docker.errors import NotFound
    try:
        size = docker_client.inspect_image(image_name).get('Size')
        if size and isinstance(size, int) and size > IMAGE_SIZE_LIMIT:
            raise shub_exceptions.CustomImageTooLargeException(
                IMAGE_TOO_LARGE_WARNING)
    except NotFound as exc:
        utils.debug_log(exc)
        raise shub_exceptions.NotFoundException(
            "The image doesn't exist yet, please use build command at first.")


def _check_shub_image_info_entry(image_name, docker_client):
    """Check that the image has shub-image-info entrypoint"""
    status, logs = _run_docker_command(
        docker_client, image_name, ['which', 'shub-image-info'])
    if status != 0 or not logs:
        _check_fallback_to_list_spiders(image_name, docker_client)


def _check_fallback_to_list_spiders(image_name, docker_client):
    status, logs = _run_docker_command(
        docker_client, image_name, ['which', 'list-spiders'])
    if status != 0 or not logs:
        raise shub_exceptions.NotFoundException(
            CONTRACT_CMD_NOT_FOUND_WARNING % 'shub-image-info (& list-spiders)')
    else:
        click.echo(LIST_SPIDERS_DEPRECATED_WARNING)


def _check_start_crawl_entry(image_name, docker_client):
    """Check that the image has start-crawl entrypoint"""
    status, logs = _run_docker_command(
        docker_client, image_name, ['which', 'start-crawl'])
    if status != 0 or not logs:
        raise shub_exceptions.NotFoundException(
            CONTRACT_CMD_NOT_FOUND_WARNING % 'start-crawl')


def _run_docker_command(client, image_name, command):
    """A helper to execute an arbitrary cmd with given docker image"""
    container = client.create_container(image=image_name, command=command)
    try:
        client.start(container)
        statuscode = client.wait(container=container['Id'])['StatusCode']
        logs = client.logs(container=container['Id'], stdout=True,
                           stderr=True if statuscode else False,
                           stream=False, timestamps=False)
        utils.debug_log(f"{command} results:\n{logs}")
        return statuscode, logs
    finally:
        client.remove_container(container)
